#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Message classes used to communicate with the :class:`IndyService`
"""

from typing import Sequence

from ..common.service import (
    ServiceAck,
    ServiceFail,
    ServiceRequest,
    ServiceResponse,
)


class IndyServiceAck(ServiceAck):
    """
    A generic acknowledgement in response to an Indy service request
    """
    pass


class IndyServiceFail(ServiceFail):
    """
    For generic errors in processing Indy requests
    """
    pass

class IndyServiceReq(ServiceRequest):
    """
    A generic Indy service request base class
    """
    pass

class IndyServiceRep(ServiceResponse):
    """
    A generic Indy service response base class
    """
    pass


class LedgerStatusReq(IndyServiceReq):
    """
    A request to fetch the status of the remote ledger
    """
    pass

class LedgerStatus(IndyServiceRep):
    """
    The response to a ledger status request
    """
    _fields = (
        "status",
    )


class RegisterWalletReq(IndyServiceReq):
    """
    A request to register a wallet
    """
    _fields = (
        ("config", dict),
    )

class WalletStatusReq(IndyServiceReq):
    """
    A request for a wallet status update
    """
    _fields = (
        ("wallet_id", str),
    )

class WalletStatus(IndyServiceRep):
    """
    A wallet status update
    """
    _fields = (
        ("wallet_id", str),
        ("status", dict),
    )


class RegisterAgentReq(IndyServiceReq):
    """
    A request to register an agent
    """
    _fields = (
        ("agent_type", str),
        ("wallet_id", str),
        ("config", dict),
    )

class AgentStatusReq(IndyServiceReq):
    """
    A request for an agent status update
    """
    _fields = (
        ("agent_id", str),
    )

class AgentStatus(IndyServiceRep):
    """
    An agent status update
    """
    _fields = (
        ("agent_id", str),
        ("status", dict),
    )


class RegisterCredentialTypeReq(IndyServiceReq):
    """
    A request to register a schema for publishing
    """
    _fields = (
        ("issuer_id", str),
        ("schema_name", str),
        ("schema_version", str),
        ("origin_did", str),
        ("attr_names", Sequence),
        ("config", dict),
    )


class RegisterConnectionReq(IndyServiceReq):
    """
    A request to register a connection
    """
    _fields = (
        ("connection_type", str),
        ("agent_id", str),
        ("config", dict),
    )

class ConnectionStatusReq(IndyServiceReq):
    """
    A request for a connection status update
    """
    _fields = (
        ("connection_id", str),
    )

class ConnectionStatus(IndyServiceRep):
    """
    A connection status update
    """
    _fields = (
        ("connection_id", str),
        ("status", dict),
    )


class IssueCredentialReq(IndyServiceReq):
    """
    Issue a credential via a previously-registered connection
    """
    _fields = (
        ("connection_id", str),
        ("schema_name", str),
        ("schema_version", str),
        ("origin_did", str),
        ("cred_data", dict),
    )


class IssueCredentialBatchReq(IndyServiceReq):
    """
    Issue a credential via a previously-registered connection
    """
    _fields = (
        ("connection_id", str),
        ("schema_name", str),
        ("schema_version", str),
        ("origin_did", str),
        ("cred_data", Sequence),
    )


class CredentialOffer(IndyServiceRep):
    """
    A successful credential offer response
    Args:
        data (dict): the resulting credential offer
    """
    _fields = (
        ("data", dict),
        ("cred_def_id", str),
    )


class CredentialRequest(IndyServiceRep):
    """
    A successful credential request response
    Args:
        cred_offer (CredentialOffer): the credential offer used as a basis
        data (str): the resulting credential request
        metadata (dict): the credential request metadata
    """
    _fields = (
        ("cred_offer", CredentialOffer),
        ("data", str),
        ("metadata", dict),
    )


class Credential(IndyServiceRep):
    """
    A successful credential creation
    """
    _fields = (
        ("cred_data", dict),
        ("cred_req_metadata", dict),
        ("cred_revoc_id", str),
    )


class StoredCredential(IndyServiceRep):
    """
    A successful response to storing a credential
    """
    _fields = (
        ("cred", Credential),
        ("cred_id", str),
        ("served_by", str, None),
    )


class StoredCredentialBatch(IndyServiceRep):
    """
    A successful response to storing a credential batch
    """
    _fields = (
        ("results", Sequence), # Sequence[StoredCredential]
        ("errors", Sequence), # Sequence[str]
    )


class GenerateCredentialRequestReq(IndyServiceReq):
    """
    A request to generate a credential request
    """
    _fields = (
        ("holder_id", str),
        ("cred_offer", CredentialOffer),
    )


class StoreCredentialReq(IndyServiceReq):
    """
    A request to store a new credential
    """
    _fields = (
        ("holder_id", str),
        ("credential", Credential),
    )


class ResolveSchemaReq(IndyServiceReq):
    """
    A request to resolve a schema which may be defined by one of our issuers
    """
    _fields = (
        ("schema_name", str),
        ("schema_version", str),
        ("origin_did", str),
    )


class ResolvedSchema(IndyServiceRep):
    """
    A request to resolve a schema which may be defined by one of our issuers
    """
    _fields = (
        ("issuer_id", str),
        ("schema_id", str),
        ("schema_name", str),
        ("schema_version", str),
        ("origin_did", str),
        ("attr_names", Sequence),
    )


class ProofRequest(IndyServiceRep):
    """
    A message representing an Indy proof request
    """
    _fields = (
        ("data", dict),
        ("wql_filters", dict, None),
    )


class ConstructProofReq(IndyServiceReq):
    """
    A request to construct a proof from a proof request
    """
    _fields = (
        ("holder_id", str),
        ("proof_req", ProofRequest),
        ("cred_ids", set),
    )


class ConstructedProof(IndyServiceRep):
    """
    A successfully constructed proof
    """
    _fields = (
        ("proof", dict),
    )


class RegisterProofSpecReq(IndyServiceReq):
    """
    A request to register a proof request specification
    """
    _fields = (
        ("config", dict),
    )


class ProofSpecStatus(IndyServiceRep):
    """
    The proof specification status update
    """
    _fields = (
        ("spec_id", str),
        ("status", dict),
    )


class GenerateProofRequestReq(IndyServiceReq):
    """
    A request to generate a proof request
    """
    _fields = (
        ("spec_id", str),
        ("wql_filters", dict, None),
    )


class RequestProofReq(IndyServiceReq):
    """
    A request to get a proof from a connection
    """
    _fields = (
        ("connection_id", str),
        ("proof_req", ProofRequest),
        ("cred_ids", set),
        ("params", dict),
    )


class VerifyProofReq(IndyServiceReq):
    """
    The message class representing a request to verify a proof
    """
    _fields = (
        ("verifier_id", str),
        ("proof_req", ProofRequest),
        ("proof", ConstructedProof),
    )


class VerifiedProof(IndyServiceRep):
    """
    The message class representing a successful proof verification
    """
    _fields = (
        ("verified", str),
        ("parsed_proof", dict),
        ("proof", ConstructedProof),
    )

class ResolveNymReq(IndyServiceReq):
    """
    The message class representing a request to resolve a DID
    """
    _fields = (
        ("did", str),
        ("agent_id", str, None),
    )

class ResolvedNym(IndyServiceRep):
    """
    The message class representing a DID resolution response
    """
    _fields = (
        ("did", str),
        ("data", dict),
    )



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Basic implementation and message classes for services running on the exchange
"""

import asyncio
import logging
from typing import Mapping

from .exchange import (
    Exchange,
    ExchangeFail,
    ExchangeMessage,
    MessageWrapper,
    RequestExecutor)
from .util import Stats

LOGGER = logging.getLogger(__name__)


class ServiceRequest(ExchangeMessage):
    """
    A standard base class for requests to a service
    """
    pass

class ServiceResponse(ExchangeMessage):
    """
    A standard base class for responses from a service
    """
    pass

class ServiceAck(ServiceResponse):
    """
    A simple acknowledgment response
    """
    pass

class ServiceFail(ExchangeFail):
    """
    A standard base class for errors returned from a service
    """
    pass

class ServiceStatusReq(ServiceRequest):
    """
    Request the status of a service
    """
    pass

class ServiceStatus(ServiceResponse):
    """
    Request the status of a service
    """
    _fields = (
        ("status", dict),
    )

class ServiceStopReq(ServiceRequest):
    """
    Request a service to stop running
    """
    pass

class ServiceSyncReq(ServiceRequest):
    """
    Request a service to perform a sync
    """
    _fields = (
        ("wait", bool),
    )

class ServiceSyncError(Exception):
    """
    An exception raised in response to a controlled failure during synchronization
    """
    pass


class ServiceBase(RequestExecutor):
    """
    The base class for services handled by the :class:`ServiceManager` instance
    """

    def __init__(self, pid: str, exchange: Exchange, env: Mapping):
        super(ServiceBase, self).__init__(pid, exchange)
        self._env = env
        self._status = {
            "id": self._pid,
            "failed": False,
            "synced": False,
            "syncing": False,
            "started": False,
        }
        self._stats = Stats()
        self._sync_again = False
        self._sync_lock = None

    def start(self, wait: bool = True) -> None:
        """
        Start the processing thread and any related services
        """
        super(ServiceBase, self).start(True)
        self._sync_lock = asyncio.Lock(loop=self._runner.loop)
        self.run_task(self._start())

    def _update_status(self, **params) -> None:
        self._status.update(params)

    async def _start(self) -> None:
        """
        Initial service startup
        """
        if await self._service_start():
            self._update_status(started=True)
            LOGGER.info("Started service: %s", self.pid)
            self.run_task(self._sync())

    async def _service_start(self) -> bool:
        """
        Perform service-specific startup actions
        """
        return True

    def send_stop_message(self) -> bool:
        return self.send_noreply(self._pid, ServiceStopReq())

    async def _stop(self) -> None:
        """
        Service shutdown
        """
        async with self._sync_lock:
            await self._service_stop()
            self._update_status(started=False)
        LOGGER.info("Stopped service: %s", self.pid)

    async def _service_stop(self) -> None:
        """
        Perform service-specific shutdown actions
        """
        pass

    async def _sync(self) -> None:
        """
        Service sync process
        """
        #pylint: disable=broad-except
        async with self._sync_lock:
            if not self._status["started"] or self._status["failed"]:
                return
            prev = self._status["synced"]
            if not prev:
                LOGGER.info("Starting sync: %s", self.pid)
            again = True
            failed = False
            synced = False
            while again:
                self._sync_again = again = False
                self._update_status(syncing=True)
                try:
                    synced = await self._service_sync()
                    if self._sync_again:
                        synced = False
                        again = True
                except ServiceSyncError:
                    LOGGER.exception("Error during %s sync: ", self.pid)
                    synced = False
                except Exception:
                    LOGGER.exception("Fatal error during %s sync: ", self.pid)
                    synced = False
                    failed = True
                self._update_status(synced=synced, syncing=False, failed=failed)
            if synced and not prev:
                LOGGER.info("Completed sync: %s", self.pid)

    def _sync_required(self) -> None:
        self._sync_again = True
        self._update_status(synced=False)

    async def _service_sync(self) -> bool:
        """
        Perform service-specific sync actions. This may be called multiple times,
        and should not repeat sync actions that aren't necessary
        """
        return True

    async def _get_status(self) -> ServiceResponse:
        """
        Return the current status of the service
        """
        result = self._status.copy()
        result["stats"] = self._stats.results()
        return ServiceStatus(result)

    async def _handle_message(self, received: MessageWrapper) -> bool:
        """
        Process a message from the exchange and send the reply, if any

        Args:
            received: The message to be processed
        """
        #pylint: disable=broad-except
        from_pid, request, ident = (
            received.from_pid,
            received.message,
            received.ident,
        )

        if await super(ServiceBase, self)._handle_message(received):
            return True

        elif isinstance(request, ServiceStopReq):
            # run service shutdown in async thread
            await self._stop()
            # finish polling
            super(ServiceBase, self).send_stop_message()
            return True

        elif isinstance(request, ServiceSyncReq):
            if request.wait:
                while True:
                    if self._status["failed"]:
                        reply = ServiceFail("Service could not be synced: {}".format(self.pid))
                        break
                    await self._sync()
                    if self._status["synced"]:
                        reply = ServiceAck()
                        break
                    await asyncio.sleep(2)
            else:
                self.run_task(self._sync())
                reply = ServiceAck()

        elif isinstance(request, ServiceStatusReq):
            reply = await self._get_status()

        elif isinstance(request, ServiceRequest):
            try:
                reply = await self._service_request(request)
            except Exception:
                LOGGER.exception("Exception while handling request:")
                reply = ServiceFail("Exception while handling request")
            if reply is None:
                raise ValueError(
                    "Unexpected message from {}: {}".format(from_pid, request)
                )

        elif isinstance(request, ServiceResponse):
            ok = await self._service_response(request)
            if not ok:
                raise ValueError(
                    "Unexpected message from {}: {}".format(from_pid, request)
                )
            return True

        self.send_noreply(from_pid, reply, ident)
        return True

    async def _service_request(self, request: ServiceRequest) -> ServiceResponse:
        """
        Handle a request from another service
        """
        pass

    async def _service_response(self, response: ServiceResponse) -> bool:
        """
        Handle a response from another service
        """
        pass

    def _timer(self, *tasks, log_as=None):
        """
        Start a new timer for a set of tasks
        """
        return self._stats.timer(*tasks, log_as=log_as)



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
:class:`IndyClient` handles message passing to the :class:`IndyService`, providing
a more natural interface for working with the ledger.
"""

from typing import Sequence

from ..common.exchange import RequestTarget
from ..common.service import (
    ServiceAck,
    ServiceRequest,
    ServiceSyncReq,
    ServiceStatusReq,
    ServiceStatus,
)

from .config import AgentType, ConnectionType
from .errors import IndyClientError

from .messages import (
    IndyServiceAck,
    IndyServiceFail,
    LedgerStatusReq,
    LedgerStatus,
    RegisterWalletReq,
    WalletStatusReq,
    WalletStatus,
    RegisterAgentReq,
    AgentStatusReq,
    AgentStatus,
    RegisterCredentialTypeReq,
    RegisterConnectionReq,
    ConnectionStatusReq,
    ConnectionStatus,
    IssueCredentialReq,
    IssueCredentialBatchReq,
    Credential,
    CredentialOffer,
    CredentialRequest,
    StoredCredential,
    StoredCredentialBatch,
    GenerateCredentialRequestReq,
    StoreCredentialReq,
    ResolveSchemaReq,
    ResolvedSchema,
    ProofRequest,
    ConstructProofReq,
    ConstructedProof,
    GenerateProofRequestReq,
    RegisterProofSpecReq,
    ProofSpecStatus,
    VerifiedProof,
    RequestProofReq,
    VerifyProofReq,
    ResolveNymReq,
    ResolvedNym,
)

class IndyClient:
    """
    This class provides a nicer interface for passing messages to the Indy service manager
    """
    def __init__(self, target: RequestTarget):
        self._target = target

    async def _fetch(self, request: ServiceRequest, expect=None):
        """
        Send a request to the :class:`IndyService` and check the result

        Args:
            request: the request to be sent
            expect: the type or types expected in response
        """
        result = await self._target.request(request)
        if isinstance(result, IndyServiceFail):
            raise IndyClientError(result.value)
        elif expect and not isinstance(result, expect):
            raise IndyClientError("Unexpected result: {}".format(result))
        return result

    async def get_ledger_status(self):
        """
        Get the status of the remote ledger (for von-network)
        """
        result = await self._fetch(LedgerStatusReq(), LedgerStatus)
        return result.status

    async def register_wallet(self, config: dict) -> str:
        """
        Register a wallet

        Args:
            config: the wallet configuration (must include 'name' and 'seed')
        Returns:
            the registered identifier of the wallet
        """
        result = await self._fetch(RegisterWalletReq(config), WalletStatus)
        return result.wallet_id

    async def get_wallet_status(self, wallet_id: str) -> dict:
        """
        Get the status of a registered wallet

        Args:
            wallet_id: the registered wallet identifier
        """
        result = await self._fetch(WalletStatusReq(wallet_id), WalletStatus)
        return result.status

    async def register_issuer(self, wallet_id: str, config: dict) -> str:
        """
        Register an issuer service

        Args:
            wallet_id: the registered wallet identifier to use for this agent
            config: the issuer configuration
        Returns:
            the registered identifier of the issuer service
        """
        agent_type = AgentType.issuer.value
        if config and config.get("holder_verifier"):
            agent_type = AgentType.combined.value
        result = await self._fetch(
            RegisterAgentReq(agent_type, wallet_id, config),
            AgentStatus)
        return result.agent_id

    async def register_holder(self, wallet_id: str, config: dict) -> str:
        """
        Register a holder service

        Args:
            wallet_id: the registered wallet identifier to use for this agent
            config: the holder configuration
        Returns:
            the registered identifier of the holder service
        """
        result = await self._fetch(
            RegisterAgentReq(AgentType.holder.value, wallet_id, config),
            AgentStatus)
        return result.agent_id

    async def register_verifier(self, wallet_id: str, config: dict) -> str:
        """
        Register a verifier service

        Args:
            wallet_id: the registered wallet identifier to use for this agent
            config: the verifier configuration
        Returns:
            the registered identifier of the verifier service
        """
        result = await self._fetch(
            RegisterAgentReq(AgentType.verifier.value, wallet_id, config),
            AgentStatus)
        return result.agent_id

    async def get_agent_status(self, agent_id: str) -> dict:
        """
        Fetch the status of a registered agent (issuer, holder, or verifier)

        Args:
            agent_id: the registered agent identifier
        """
        result = await self._fetch(AgentStatusReq(agent_id), AgentStatus)
        return result.status

    async def register_credential_type(self, issuer_id: str,
                                       schema_name: str, schema_version: str,
                                       origin_did: str, attr_names: Sequence,
                                       config: dict = None) -> None:
        """
        Register a credential type for a previously-registered issuer

        Args:
            issuer_id: the registered agent identifier
            schema_name: the name of the schema
            schema_version: the version of the schema
            origin_did: for schemas published by other issuers, otherwise None
            attr_names: the list of attribute names, required for a schema to be published
            config: extra configuration parameters for the credential type
        """
        await self._fetch(
            RegisterCredentialTypeReq(
                issuer_id, schema_name, schema_version,
                origin_did, attr_names, config),
            IndyServiceAck)

    async def register_http_connection(self, agent_id: str, config: dict = None) -> str:
        """
        Register an HTTP connection to a holder/prover service

        Args:
            agent_id: the registered issuer or verifier agent identifier
            config: configuration parameters for the connection (must include 'api_url')
        """
        result = await self._fetch(
            RegisterConnectionReq(ConnectionType.HTTP.value, agent_id, config or {}),
            ConnectionStatus)
        return result.connection_id

    async def register_orgbook_connection(self, agent_id: str, config: dict = None) -> str:
        """
        Register a connection to TheOrgBook as a holder/prover

        Args:
            agent_id: the registered issuer or verifier agent identifier
            config: configuration parameters for the connection (must include 'api_url')
        """
        result = await self._fetch(
            RegisterConnectionReq(ConnectionType.TheOrgBook.value, agent_id, config or {}),
            ConnectionStatus)
        return result.connection_id

    async def register_holder_connection(self, agent_id: str, config: dict = None) -> str:
        """
        Register a connection to a local holder agent

        Args:
            agent_id: the identifier of the issuer or verifier connecting to this holder
            config: extra configuration parameters for the connection (must include 'holder_id')
        """
        result = await self._fetch(
            RegisterConnectionReq(ConnectionType.holder.value, agent_id, config or {}),
            ConnectionStatus)
        return result.connection_id

    async def get_connection_status(self, connection_id: str) -> dict:
        """
        Fetch the status of a registered connection

        Args:
            connection_id: the registered connection identifier
        """
        result = await self._fetch(ConnectionStatusReq(connection_id), ConnectionStatus)
        return result.status

    async def issue_credential(self, connection_id: str, schema_name: str, schema_version: str,
                               origin_did: str, cred_data: dict) -> StoredCredential:
        """
        Issue a credential to a previously-registered connection

        Args:
            connection_id: the registered connection identifier
            schema_name: the name of the schema, used to identify the credential type
            schema_version: the version of the schema
            origin_did: the origin DID of the schema, if external
            cred_data: the new credential's raw claim attribute values
        """
        return await self._fetch(
            IssueCredentialReq(
                connection_id, schema_name, schema_version, origin_did, cred_data),
            StoredCredential)

    async def issue_credential_batch(
            self, connection_id: str, schema_name: str, schema_version: str,
            origin_did: str, cred_data: Sequence[dict]) -> StoredCredentialBatch:
        """
        Issue a list of credentials to a previously-registered connection

        Args:
            connection_id: the registered connection identifier
            schema_name: the name of the schema, used to identify the credential type
            schema_version: the version of the schema
            origin_did: the origin DID of the schema, if external
            cred_data: the list of new credential's raw claim attribute values
        """
        return await self._fetch(
            IssueCredentialBatchReq(
                connection_id, schema_name, schema_version, origin_did, cred_data),
            StoredCredentialBatch)

    async def create_credential_request(self, holder_id: str, cred_offer: dict,
                                        cred_def_id: str) -> CredentialRequest:
        """
        Create a credential request for an issuer service

        Args:
            holder_id: the registered agent identifier of the holder service
            cred_offer: the Indy credential offer received from the issuer
            cred_def_id: The identifier of the credential definition
        """
        return await self._fetch(
            GenerateCredentialRequestReq(holder_id, CredentialOffer(cred_offer, cred_def_id)),
            CredentialRequest)

    async def store_credential(self, holder_id: str,
                               credential: Credential) -> StoredCredential:
        """
        Store a credential in a holder's wallet

        Args:
            holder_id: the registered agent identifier
            credential: the Indy credential record
        """
        return await self._fetch(
            StoreCredentialReq(holder_id, credential),
            StoredCredential)

    async def resolve_schema(self, name: str, version: str = None,
                             origin_did: str = None) -> ResolvedSchema:
        """
        Resolve a schema from a registered credential type or the ledger

        Args:
            name: the schema name
            version: the schema version
            origin_did: the DID of the schema issuer
        """
        return await self._fetch(
            ResolveSchemaReq(name, version, origin_did),
            ResolvedSchema)

    async def construct_proof(self, holder_id: str, proof_req: dict,
                              wql_filters: dict = None, cred_ids: set = None) -> ConstructedProof:
        """
        Construct a proof from credentials in the holder's wallet given a proof request

        Args:
            holder_id: the registered agent identifier
            proof_req: the Indy proof request record
            cred_ids: an optional set of credential IDs to use in the proof
        """
        return await self._fetch(
            ConstructProofReq(holder_id, ProofRequest(proof_req, wql_filters), cred_ids),
            ConstructedProof)

    async def register_proof_spec(self, spec: dict) -> str:
        """
        Register a proof request specification

        Args:
            spec: the proof request specification
        Returns:
            the identifier of the registered proof request spec
        """
        result = await self._fetch(
            RegisterProofSpecReq(spec),
            ProofSpecStatus)
        return result.spec_id

    async def generate_proof_request(self, spec_id: str) -> ProofRequest:
        """
        Generate a proof request based on a previously-registered proof request spec

        Args:
            spec_id: the registered proof request spec identifier
        """
        return await self._fetch(
            GenerateProofRequestReq(spec_id),
            ProofRequest)

    async def request_proof(self, connection_id: str, proof_req: ProofRequest,
                            cred_ids: set = None, params: dict = None) -> ConstructedProof:
        """
        Request a proof from a holder connection

        Args:
            connection_id: the registered holder connection
            proof_req: the Indy proof request record
            cred_ids: an optional set of credential IDs to use in the proof
            params: extra parameters for the connection to use in the proof construction
        """
        return await self._fetch(
            RequestProofReq(connection_id, proof_req, cred_ids, params),
            VerifiedProof)

    async def verify_proof(self, verifier_id: str, proof_req: ProofRequest,
                           proof: ConstructedProof) -> VerifiedProof:
        """
        Verify a previously constructed proof

        Args:
            verifier_id: the registered verifier agent
            proof_req: the Indy proof request record
            proof: the constructed proof
        """
        return await self._fetch(
            VerifyProofReq(verifier_id, proof_req, proof),
            VerifiedProof)

    async def resolve_nym(self, did: str, agent_id: str = None):
        """
        Resolve a DID on the ledger

        Args:
            did: the DID to resolve
            agent_id: the agent instance to use
        """
        return await self._fetch(
            ResolveNymReq(did, agent_id),
            ResolvedNym)

    async def sync(self, wait: bool = True) -> bool:
        """
        Request the :class:`IndyService` to perform synchronization of registered services

        Args:
            wait: whether to return immediately or wait for the sync to finish
        """
        result = await self._fetch(
            ServiceSyncReq(wait))
        if isinstance(result, ServiceAck):
            return True
        return False

    async def get_status(self) -> dict:
        """
        Fetch the status of the :class:`IndyService` and its registered services
        """
        result = await self._fetch(
            ServiceStatusReq(),
            ServiceStatus)
        return result.status



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Handle processing of issue-credential forms
"""

import logging

from aiohttp import web

from . import helpers
from ..indy.errors import IndyClientError

LOGGER = logging.getLogger(__name__)


def load_cred_request(form, attr_names, request: web.Request) -> dict:
    """
    Create a new credential request from a `issue-credential` form definition, fetching
    input from the client request as necessary
    """

    cred = {}
    mapping = form.get("mapping") or {}
    if mapping.get("fill_defaults", True):
        for attr in attr_names:
            cred[attr] = request.get(attr)
            LOGGER.debug("credential %s %s", attr, cred[attr])
    map_attr = mapping.get("attributes") or []
    # Build credential data from schema mapping
    for attribute in map_attr:
        attr_name = attribute.get("name")
        from_type = attribute.get("from", "request")
        # Handle getting value from request data
        if from_type == "request":
            source = attribute.get("source", attr_name)
            cred[attr_name] = request.get(source)
        # Handle getting value from helpers (function defined in config)
        elif from_type == "helper":
            helper = getattr(helpers, attribute["source"], None)
            if not helper:
                raise ValueError("Cannot find helper '{}'".format(attribute["source"]))
            cred[attribute["name"]] = helper()
        # Handle setting value with string literal or None
        elif from_type == "literal":
            cred[attr_name] = attribute.get("source")
        # Handle getting value already set on schema skeleton
        elif from_type == "previous":
            source = attribute.get("source")
            if source:
                try:
                    cred[attr_name] = cred[source]
                except KeyError:
                    raise ValueError(
                        "Cannot find previous value '{}'".format(source))
        else:
            raise ValueError("Unknown mapping type '{}'".format(attribute["from"]))
    return cred


async def process_form(form, request: web.Request) -> web.Response:
    """
    Handle `issue-credential` form processing by looking up the issuer for the schema
    and issuing the credential
    """

    #pylint: disable=broad-except
    if form["type"] == "issue-credential":
        schema_name = form.get("schema_name")
        schema_version = form.get("schema_version")
        if not schema_name:
            # FIXME should be an internal error
            return web.Response(reason="Form definition missing 'schema_name'", status=400)

        LOGGER.debug("request %s", request)
        inputs = await request.json()
        if isinstance(inputs, dict):
            inputs = inputs.get("attributes") or {}
        else:
            inputs = await request.post()

        client = request.app["manager"].get_client()
        try:
            result = await client.resolve_schema(schema_name, schema_version)
        except IndyClientError:
            msg = "Issuer for schema '{}' is not defined or not loaded".format(schema_name)
            return web.Response(reason=msg, status=400)

        params = load_cred_request(form, result.attr_names, inputs)
        #return web.json_response(params)

        try:
            stored = await client.issue_credential(
                result.issuer_id, result.schema_name, result.schema_version,
                result.origin_did, params)
        except IndyClientError as e:
            ret = {"success": False, "result": str(e)}
        else:
            ret = {"success": True, "result": stored.cred_id}

        #if ret["success"]:
        #    return response.html('<h3>Registration successful</h3>')
        #else:
        #    return response.html('<h3>Registration could not be completed</h3>')
        return web.json_response(ret)
    return web.Response(reason="Method not supported", status=405)



import asyncio
import logging
import os
import sys

from vonx.indy.manager import IndyManager

LOGGER = logging.getLogger()


class TestIndyIssuer(IndyManager):
    """
    A test Indy service manager which creates sample wallets and issuers
    """

    schema_name = "test.schema"
    schema_version = "1.0.0"


    def get_service_init_params(self) -> dict:
        return {
            "auto_register": 1,
            "genesis_path": "/home/indy/genesis",
            "ledger_url": os.environ.get("LEDGER_URL") or "http://192.168.65.3:9000",
        }

    async def _load_config(self) -> None:
        # skip default implementation
        pass

    async def add_test_services(self, client):
        LOGGER.info("setting up test indy issuer")
        holder_url = os.environ.get("HOLDER_API") or "http://192.168.65.3:5001/holder"

        all = {}

        all["issuer_wallet_id"] = await client.register_wallet({
            "name": "issuer-wallet",
            "seed": "issuer-wallet-000000000000000001",
        })
        all["issuer_id"] = await client.register_issuer(all["issuer_wallet_id"], {
            "name": "Test Issuer",
            "email": "test@example.ca",
        })
        mapping = [
            {
                "model": "name",
                "fields": {
                    "text": {
                        "input": "attr1",
                        "from": "claim"
                    },
                    "type": {
                        "input": "legal_name",
                        "from": "value"
                    }
                }
            }
        ]
        await client.register_credential_type(
            all["issuer_id"],
            self.schema_name,
            self.schema_version,
            None,
            ["attr1", "attr2"],
            {
                "description": "Test Credential",
                "source_claim": "attr1",
                "mapping": mapping,
            }
        )

        all["holder_conn_id"] = await client.register_http_connection(
            all["issuer_id"], {
                "id": "holder",
                "api_url": holder_url,
            }
        )

        all["verifier_wallet_id"] = await client.register_wallet({
            "name": "verifier-wallet",
            "seed": "verifier-wallet-0000000000000001",
        })
        all["verifier_id"] = await client.register_verifier(all["verifier_wallet_id"], {
            "name": "Test Verifier",
        })
        all["verifier_conn_id"] = await client.register_http_connection(
            all["verifier_id"], {
                "id": "verifier",
                "api_url": holder_url,
            }
        )
        proof_spec = {
            "id": "test-proof",
            "version": "1.0.0",
            "schemas": [
                {
                    "key": {
                        # "did": "PXocv6sBRa7YefPvnHpsqp",
                        "name": self.schema_name,
                        "version": self.schema_version,
                    },
                    #"predicates": [
                    #    {
                    #        "name": "attr1",
                    #        "p_type": ">=",
                    #        "p_value": 1,
                    #    }
                    #]
                }
            ]
        }
        all["proof_spec_id"] = await client.register_proof_spec(proof_spec)

        return all


    async def test_issue_creds(self, client, conn_id, count):
        LOGGER.info("--- issuing %s test credentials ---", count)
        creds = [self.test_issue_cred(client, conn_id)
                 for _ in range(count)]
        start = time.time()
        await asyncio.gather(*creds)
        dur = time.time() - start
        avg = dur / len(creds)
        LOGGER.info("--- issued %s creds in %s seconds, avg %s ---", len(creds), dur, avg)


    async def test_issue_cred(self, client, conn_id) -> str:
        stored = await client.issue_credential(
            conn_id, self.schema_name, self.schema_version, None,
            {"attr1": "Test", "attr2": "Second Value"})
        LOGGER.info("issued: %s", stored.cred_id)
        return stored.cred_id


    async def test_proof(self, client, conn_id, spec_id, cred_ids=None):
        proof_req = await client.generate_proof_request(spec_id)
        #cred_ids = None
        result = await client.request_proof(conn_id, proof_req, cred_ids)
        LOGGER.info("test proof verified: %s, result: %s", result.verified, result.parsed_proof)


def test_web(manager):
    from vonx.web import init_web
    app = init_web(manager)

    from aiohttp import web
    web.run_app(app, host='0.0.0.0', port=5000)


if __name__ == '__main__':
    LOGGER.setLevel(logging.DEBUG)
    CONSOLE = logging.StreamHandler()
    CONSOLE.setLevel(os.environ.get("LOG_LEVEL", logging.INFO))
    LOGGER.addHandler(CONSOLE)

    MGR = TestIndyIssuer()
    MGR.start()

    TEST = sys.argv[1] if len(sys.argv) > 1 else "proof"

    CLIENT = MGR.get_client()
    DONE = False
    async def setup(client, teardown=False):
        ids = await MGR.add_test_services(client)
        if TEST == "proof" and await client.sync():
            cred_id = await MGR.test_issue_cred(client, ids["holder_conn_id"])
            await MGR.test_proof(client, ids["verifier_conn_id"], ids["proof_spec_id"], {cred_id})
        if teardown:
            MGR.stop()

    try:
        asyncio.get_event_loop().run_until_complete(setup(CLIENT, TEST != "web"))

        if TEST == "web":
            test_web(MGR)
    except:
        LOGGER.exception("Error during init")
        MGR.stop()

    LOGGER.info("done")



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Methods and classes for working with asyncio event loops
"""

import asyncio
from concurrent.futures import Executor, Future
from threading import get_ident, Event, Thread
from typing import Awaitable, Callable, Coroutine
import logging

LOGGER = logging.getLogger(__name__)


def run_coro(coro: Coroutine):
    """
    Run an async coroutine and wait for the results

    Args:
        coro (CoroutineType): The coroutine to execute
    Returns:
        The result of the coroutine
    """
    event_loop = None
    try:
        event_loop = asyncio.get_event_loop()
    except RuntimeError:
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
    return event_loop.run_until_complete(coro)


def run_in_executor(executor: Executor, coro: Coroutine) -> Future:
    """
    Run an async coroutine in an executor when we aren't already inside an event loop

    Args:
        executor: A :class:`ThreadExecutor` or :class:`ProcessExecutor` instance which will
            run the coroutine
    Returns:
        A `Future` which can be used to access the result of the coroutine
    """
    loop = asyncio.new_event_loop()
    def _run_sync_loop(loop, coro):
        asyncio.set_event_loop(loop)
        loop.run_until_complete(coro)
    future = executor.submit(_run_sync_loop, loop, coro)
    return future


class Runner:
    """
    Run a new event loop in a separate thread and allow tasks to be submitted to it
    """
    def __init__(self, loop=None):
        self._active = False
        self._loop = loop
        self._thread = None

    @property
    def loop(self):
        """
        Accessor for the event loop instance
        """
        return self._loop

    def start(self, wait: bool = True) -> None:
        """
        Run the event loop in a new thread

        Args:
            wait: block until the event loop is running
        """
        if self._active:
            return
        if not self._loop:
            self._loop = asyncio.new_event_loop()
        event = Event() if wait else None
        self._thread = Thread(target=self._run, args=(event,))
        self._thread.daemon = True
        self._thread.start()
        if event:
            event.wait()

    def _run(self, event=None) -> None:
        """
        The main logic of the event loop thread
        """
        asyncio.set_event_loop(self._loop)
        def _ready():
            self._active = True
            if event:
                event.set()
        self._loop.call_soon(_ready)
        self._loop.run_forever()

    def stop(self, wait: bool = True) -> None:
        """
        Terminate the event loop thread

        Args:
            wait: block until the event loop has been stopped
        """
        def _finish(event):
            self._active = False
            self._loop.stop()
            #pending = asyncio.Task.all_tasks()
            #self._loop.run_until_complete(asyncio.gather(*pending))
            if event:
                event.set()
        if get_ident() == self._thread.ident:
            _finish(None)
        else:
            event = Event() if wait else None
            self._loop.call_soon_threadsafe(_finish, event)
            if event:
                event.wait()
                self.join()

    def join(self):
        """
        Wait for the event loop thread to terminate
        """
        return self._thread.join()

    def _add_task(self, coro: Awaitable, future: Future = None) -> asyncio.Future:
        """
        Add a coroutine to the event loop, to be run at a later time

        Args:
            coro: the coroutine to be added
            future: an optional :class:`Future` used to return the result to another thread
        """
        result = asyncio.ensure_future(coro, loop=self._loop)
        if future:
            future.set_result(result)
        return result

    def run_task(self, coro: Awaitable) -> asyncio.Future:
        """
        Add a coroutine to the event loop, to be run at a later time

        Args:
            coro: the coroutine to be added
        """
        if not self._active:
            raise RuntimeError('Runner is not active')
        if get_ident() == self._thread.ident:
            result = self._add_task(coro)
        else:
            fut = Future()
            self._loop.call_soon_threadsafe(self._add_task, coro, fut)
            result = fut.result()
        return result

    def run_in_executor(self, executor: Executor, func: Callable, *args) -> asyncio.Future:
        """
        Run a function in an executor, in the runner's event loop

        Args:
            executor: the Executor to use, may be None for the default ThreadPoolExecutor
            func: the function to run
            args: arguments to pass to the function
        """
        if not self._active:
            raise RuntimeError('Runner is not active')
        coro = self._loop.run_in_executor(executor, func, *args)
        return self.run_task(coro)



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Define routes for the aiohttp webserver application based on the defined configuration
"""

import logging
from typing import Coroutine

from aiohttp import web, ClientRequest

from ..common.manager import ConfigServiceManager
from . import views
from .process import process_form
from .render import render_form

LOGGER = logging.getLogger(__name__)


def get_standard_routes(_app) -> list:
    """
    Get the standard list of routes for the von-x application
    """
    return [
        web.get('/health', views.health),
        web.get('/status', views.status),
        web.get('/ledger-status', views.ledger_status),
        web.post('/issue-credential', views.issue_credential),
        web.post('/{connection_id}/issue-credential', views.issue_credential),
        web.post('/request-proof', views.request_proof),
        web.post('/{connection_id}/request-proof', views.request_proof),
        web.post('/{holder_id}/generate-credential-request', views.generate_credential_request),
        web.post('/{holder_id}/store-credential', views.store_credential),
        web.post('/{holder_id}/construct-proof', views.construct_proof),
        #web.get('/hello', views.hello),
    ]


def get_custom_routes(app: web.Application) -> list:
    """
    Get the list of routes defined by the application route settings
    """
    return RouteDefinitions.load(app['manager']).routes


def get_routes(app: web.Application) -> list:
    """
    Get the full list of defined routes
    """
    return get_standard_routes(app) + get_custom_routes(app)


class RouteDefinitions:
    """
    Manager class for loading and inspecting the application routing configuration
    """
    def __init__(self):
        self.forms = []
        self.paths = []

    @classmethod
    def load(cls, manager: ConfigServiceManager) -> 'RouteDefinitions':
        """
        Return a new instance initialized by a :class:`ConfigServiceManager`
        """
        inst = RouteDefinitions()
        inst.load_config(manager)
        return inst

    def add_paths(self, *paths, overwrite: bool = False) -> None:
        """
        Add a list of paths to the defined routes

        Args:
            paths: one or more paths to add
            overwrite: whether to replace an existing path
        """
        for path in paths:
            if self.path_defined(path):
                if not overwrite:
                    raise RuntimeError('Duplicate view path defined: {}'.format(path))
            else:
                self.paths.append(path)

    def add_form(self, form: dict) -> None:
        """
        Add a form route definition

        Args:
            form: a dictionary of form configuration parameters
        """
        self.add_paths(form['path'])
        self.forms.append(form)


    def path_defined(self, path: str) -> bool:
        """
        Check whether a given path is defined by a previously-added route
        """
        return path in self.paths

    def load_config(self, manager: ConfigServiceManager) -> bool:
        """
        Load the standard route configuration defined by a :class:`ConfigServiceManager` instance
        and its environment variables
        """
        config = manager.load_config_path('ROUTES_CONFIG_PATH', 'routes.yml')
        if not config:
            return False

        limit_forms = manager.env.get('FORMS')
        limit_forms = limit_forms.split() \
            if (limit_forms and limit_forms != 'all') \
            else None

        forms = config.get('forms') or {}
        self.load_form_definitions(forms, limit_forms)

        return True

    def load_form_definitions(self, config: dict, limit_forms=None) -> None:
        """
        Load a dictionary of form definitions from the application route configuration
        """
        for form_id, form in config.items():
            if limit_forms is not None and form_id not in limit_forms:
                continue
            form_id = form['id'] = form.get('id', form_id)
            if not 'name' in form:
                form['name'] = form_id
            if not form.get('path'):
                form['path'] = '/' + form['name']
            check_form_definition(form)
            self.add_form(form)

    @property
    def routes(self) -> list:
        """
        Accessor for the combined list of routes defined by our configuration
        """
        routes = []

        routes.extend(
            web.view(form['path'], form_handler(form), name=form['name'])
            for form in self.forms)

        return routes


def check_form_definition(form: dict) -> None:
    """
    Verify a form definition and expand properties as required
    """
    supported_types = ['issue-credential']

    form_id = form.get('id')
    form_type = form.get('type')
    if not form_type:
        raise ValueError('Type not defined for form: {}'.format(form_id))
    if form_type not in supported_types:
        raise ValueError('Unknown form type for {}: {}'.format(form_id, form_type))


def form_handler(form: dict) -> Coroutine:
    """
    Return a request handler for processing form routes
    """
    async def _process(request: ClientRequest):
        if request.method == 'GET' or request.method == 'HEAD':
            return await render_form(form, request)
        elif request.method == 'POST':
            return await process_form(form, request)
        return web.Response(status=405)
    return _process



#!/usr/bin/env python3
#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import argparse
import json
import os
import sys
import time

import aiohttp

DEFAULT_AGENT_URL = os.environ.get('AGENT_URL', 'http://localhost:5000')

parser = argparse.ArgumentParser(
    description='Issue one or more credentials via von-x')
parser.add_argument('conn_id', help='the connection ID')
parser.add_argument('paths', nargs='+',
    help='the path to a credential JSON file')
parser.add_argument('-c', '--count', type=int, default=1,
    help='repeatedly issue claims this many times (for volume testing)')
parser.add_argument('-p', '--parallel', action='store_true',
    help='submit the credentials in parallel')
parser.add_argument('-u', '--url', default=DEFAULT_AGENT_URL,
    help='the URL of the von-x service')

args = parser.parse_args()

AGENT_URL = args.url
CONN_ID = args.conn_id
CRED_PATHS = args.paths
PARALLEL = args.parallel
REPEAT = args.count

async def issue_cred(http_client, conn_id, cred_path, ident):
    with open(cred_path) as cred_file:
        cred = json.load(cred_file)
    if not cred:
        raise ValueError('Credential could not be parsed')
    schema = cred.get('schema')
    if not schema:
        raise ValueError('No schema defined')
    version = cred.get('version', '')
    attrs = cred.get('attributes')
    if not attrs:
        raise ValueError('No schema attributes defined')

    print('Submitting credential {} {}'.format(ident, cred_path))

    start = time.time()
    try:
        response = await http_client.post(
            '{}/issue-credential'.format(AGENT_URL),
            params={'schema': schema, 'version': version, 'connection_id': conn_id},
            json=attrs
        )
        if response.status != 200:
            raise RuntimeError(
                'Credential could not be processed: {}'.format(await response.text())
            )
        result_json = await response.json()
    except Exception as exc:
        raise Exception(
            'Could not issue credential. '
            'Are von-x and TheOrgBook running?') from exc

    elapsed = time.time() - start
    print('Response to {} from von-x ({:.2f}s):\n\n{}\n'.format(ident, elapsed, result_json))

async def submit_all(conn_id, cred_paths, parallel=False, repeat=1):
    start = time.time()
    async with aiohttp.ClientSession() as http_client:
        all = []
        idx = 1
        for cred_path in cred_paths:
            for ridx in range(args.count):
                req = issue_cred(http_client, conn_id, cred_path, idx)
                if parallel:
                    all.append(req)
                else:
                    await req
                idx += 1
        if all:
            await asyncio.gather(*all)
    elapsed = time.time() - start
    print('Total time: {:.2f}s'.format(elapsed))

asyncio.get_event_loop().run_until_complete(submit_all(CONN_ID, CRED_PATHS, PARALLEL, REPEAT))



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Abstractions to remove direct client dependency on didauth library
"""

import logging
import time
from typing import Mapping

from didauth.base import KeyFinderBase, StaticKeyFinder
from didauth.error import VerifierException
from didauth.headers import HeaderVerifier
from didauth.utils import decode_string

from vonx.indy.client import IndyClient
from vonx.indy.errors import IndyError

LOGGER = logging.getLogger(__name__)


async def verify_signature(
        headers: Mapping, key_finder: KeyFinderBase, method: str = None, path: str = None):
    """
    Verify the DID signature on request headers

    Args:
        headers: the collection of request headers
        key_finder: the key finder (likely IndyKeyFinder or KeyCache)
        method: the HTTP method
        path: the request path including the query string
    """
    verifier = HeaderVerifier(key_finder)
    try:
        return await verifier.verify(headers, method, path)
    except VerifierException as e:
        raise IndyError("Could not verify request headers") from e


class IndyKeyFinder(KeyFinderBase):
    """
    Look up the public key for an issuer
    """

    def __init__(self, client: IndyClient, verifier_id: str, source: KeyFinderBase = None):
        super(IndyKeyFinder, self).__init__(source)
        self._client = client
        self._verifier_id = verifier_id

    async def _lookup_key(self, key_id: str, key_type: str):
        if key_type != "ed25519":
            return None
        if key_id.startswith("did:sov:"):
            short_key_id = key_id[8:]
        else:
            short_key_id = key_id
            key_id = "did:sov:" + short_key_id

        LOGGER.debug("Fetching verkey for DID '%s' from ledger", key_id)
        nym_info = await self._client.resolve_nym(short_key_id, self._verifier_id)
        if nym_info.data and nym_info.data.get("verkey"):
            verkey = nym_info.data["verkey"]
            try:
                if verkey[0] == "~":
                    did = decode_string(nym_info.data["dest"], "base58")
                    suffix = decode_string(verkey[1:], "base58")
                    return did + suffix
                return decode_string(verkey, "base58")
            except ValueError:
                raise IndyError("Cannot decode verkey from ledger as base58: {}".format(nym_info.data["verkey"]))
        return None


class KeyCache(StaticKeyFinder):
    def __init__(self, source: KeyFinderBase, expiry=600):
        super(KeyCache, self).__init__(source, True)
        self._expiry = expiry
        self._updated = {}

    def add_key(self, key_id: str, key_type: str, key: bytes):
        super(KeyCache, self).add_key(key_id, key_type, key)
        if key:
            self._updated[key] = time.time()

    async def _lookup_key(self, key_id: str, key_type: str) -> bytes:
        key = await super(KeyCache, self)._lookup_key(key_id, key_type)
        if key and self._expiry and key in self._updated and \
                self._updated[key] + self._expiry < time.time():
            LOGGER.debug("Ignoring expired cache key")
            key = None
        return key


__all__ = ('IndyKeyFinder', 'KeyCache', 'KeyFinderBase', 'verify_signature')



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
vonx.web module initialization
"""


import json
import os

from aiohttp import web
import aiohttp_jinja2
from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader

from ..common.manager import ConfigServiceManager
from .routes import get_routes


def _setup_jinja(manager: ConfigServiceManager, app: web.Application):
    """
    Initialize aiohttp-jinja2 for template rendering
    """

    tpl_path = manager.env.get('TEMPLATE_PATH')
    if not tpl_path:
        tpl_path = os.path.join(manager.config_root, 'templates')
    # load default templates provided by package
    loader = PackageLoader('vonx', 'templates')
    if tpl_path:
        # load custom templates if present
        # may want to use a resource loader if tpl_path looks like a package name (has a colon)
        loader = ChoiceLoader([
            loader,
            FileSystemLoader(tpl_path)
        ])
    filters = {"jsonify": json.dumps}
    aiohttp_jinja2.setup(app, loader=loader, filters=filters)


async def init_web(manager: ConfigServiceManager):
    """
    Initialize the web server application
    """
    base = manager.env.get('WEB_BASE_HREF', '/')

    app = web.Application()
    app['base_href'] = base
    app['manager'] = manager
    app['static_root_url'] = base + 'assets'
    app.add_routes(get_routes(app))
    _setup_jinja(manager, app)

    if base != '/':
        root_app = web.Application()
        root_app.add_subapp(base, app)
        return root_app
    return app



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Methods for loading and working with our standard YAML-based configuration files
"""

import logging
import os
import re
from typing import Callable, Mapping, TextIO

import pkg_resources
import yaml


def load_resource(path: str) -> TextIO:
    """
    Open a resource file located in a python package or the local filesystem

    Args:
        path (str): The resource path in the form of `dir/file` or `package:dir/file`
    Returns:
        A file-like object representing the resource
    """
    components = path.rsplit(':', 1)
    if len(components) == 1:
        return open(components[0])
    return pkg_resources.resource_stream(components[0], components[1])


def load_settings(env=True) -> dict:
    """
    Loads the application settings from several sources:

        - settings.yml
        - an optional application settings file defined by SETTINGS_PATH
        - custom environment settings defined by ENVIRONMENT (ie. dev, prod)
        - enviroment variable overrides

    Args:
        env: A dict of environment variables, or the value True to inherit the global
            environment
    Returns:
        A combined dictionary of setting values
    """
    if env is True:
        env = os.environ
    elif not env:
        env = {}
    env_name = os.environ.get('ENVIRONMENT', 'default')

    settings = {}

    # Load default settings
    with load_resource('vonx.config:settings.yml') as resource:
        cfg = yaml.load(resource)
        if 'default' not in cfg:
            raise ValueError('Default settings not found in settings.yml')
        settings.update(cfg['default'])
        if env_name != 'default' and env_name in cfg:
            settings.update(cfg[env_name])

    # Load application settings
    ext_path = os.environ.get('SETTINGS_PATH')
    if not ext_path:
        config_root = os.environ.get('CONFIG_ROOT', os.curdir)
        ext_path = os.path.join(config_root, 'settings.yml')
    with load_resource(ext_path) as resource:
        ext_cfg = yaml.load(resource)
        if 'default' in ext_cfg:
            settings.update(ext_cfg['default'])
        if env_name != 'default':
            if env_name not in ext_cfg:
                raise ValueError(
                    'Environment not defined by application settings: {}'.format(env_name))
            settings.update(ext_cfg[env_name])

    # Inherit environment variables
    for k, v in env.items():
        if v is not None and v != '':
            settings[k] = v

    # Expand variable references
    for k, v in settings.items():
        if isinstance(v, str):
            settings[k] = expand_string_variables(v, settings)

    return settings


def load_config(path: str, env=None):
    """
    Load a YAML config file and replace variables from the environment

    Args:
        path (str): The resource path in the form of `dir/file` or `package:dir/file`
    Returns:
        The configuration tree with variable references replaced, or `False` if the
        file is not found
    """
    try:
        with load_resource(path) as resource:
            cfg = yaml.load(resource)
    except FileNotFoundError:
        return False
    cfg = expand_tree_variables(cfg, env or os.environ)
    return cfg


def expand_string_variables(value, env: Mapping, warn: bool = True):
    """
    Expand environment variables of form `$var` and `${var}` in a string

    Args:
        value (str): The input value
        env (Mapping): The dictionary of environment variables
        warn (bool): Whether to warn on references to undefined variables
    Returns:
        The transformed string
    """
    if not isinstance(value, str):
        return value
    def _replace_var(matched):
        default = None
        var = matched.group(1)
        if matched.group(2):
            var = matched.group(2)
            default = matched.group(4)
        found = env.get(var)
        if found is None or found == '':
            found = default
        if found is None and warn:
            logging.getLogger(__name__).warning('Configuration variable not defined: %s', var)
            found = ''
        return found
    return re.sub(r'\$(?:(\w+)|\{([^}]*?)(:-([^}]*))?\})', _replace_var, value)


def map_tree(tree, map_fn: Callable):
    """
    Map one tree to another using a transformation function

    Args:
        tree: A sequence, mapping or other value
        map_fn (Callable): The function to apply to each node, returing the new value
    Returns:
        The transformed tree
    """
    if isinstance(tree, Mapping):
        return {key: map_tree(value, map_fn) for (key, value) in tree.items()}
    if isinstance(tree, (list, tuple)):
        return [map_tree(value, map_fn) for value in tree]
    return map_fn(tree)


def expand_tree_variables(tree, env: Mapping, warn: bool = True):
    """
    Expand environment variables of form `$var` and `${var}` in a configuration tree.
    This is used to allow variable insertion in issuer and route definitions

    Args:
        tree: A sequence, mapping or other value
        env (Mapping): The dictionary of environment variables
        warn (bool): Whether to warn on references to undefined variables
    Returns:
        The transformed tree
    """
    return map_tree(tree, lambda val: expand_string_variables(val, env, warn))



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
The Indy service implements handlers for all the ledger-related messages, sychronizes
agents and connections, and handles the core logic for working with credentials and proofs.
"""

import asyncio
import base64
import json
import hashlib
import logging
import pathlib
import random
import string
import time
from typing import Mapping, Sequence

from didauth.ext.aiohttp import SignedRequest, SignedRequestAuth
from von_anchor.error import AbsentCred, AbsentSchema, AbsentCredDef
from von_anchor.nodepool import NodePool
from von_anchor.util import cred_def_id, revealed_attrs, schema_id, schema_key, \
    proof_req_infos2briefs, proof_req_briefs2req_creds

from ..common.service import (
    Exchange,
    ServiceBase,
    ServiceRequest,
    ServiceResponse,
    ServiceSyncError,
)
from ..common.util import log_json
from .config import (
    AgentType,
    AgentCfg,
    ConnectionCfg,
    ProofSpecCfg,
    SchemaCfg,
    WalletCfg,
)
from .connection import HttpSession
from .errors import IndyConfigError, IndyConnectionError, IndyError
from .messages import (
    IndyServiceAck,
    IndyServiceFail,
    LedgerStatusReq,
    LedgerStatus,
    RegisterWalletReq,
    WalletStatusReq,
    WalletStatus,
    RegisterAgentReq,
    AgentStatusReq,
    AgentStatus,
    RegisterCredentialTypeReq,
    RegisterConnectionReq,
    ConnectionStatusReq,
    ConnectionStatus,
    IssueCredentialReq,
    IssueCredentialBatchReq,
    Credential,
    CredentialOffer,
    CredentialRequest,
    StoredCredential,
    GenerateCredentialRequestReq,
    StoreCredentialReq,
    ResolveSchemaReq,
    ResolvedSchema,
    ProofRequest,
    ConstructProofReq,
    ConstructedProof,
    RegisterProofSpecReq,
    ProofSpecStatus,
    GenerateProofRequestReq,
    RequestProofReq,
    VerifiedProof,
    VerifyProofReq,
    ResolveNymReq,
    ResolvedNym,
)

LOGGER = logging.getLogger(__name__)


def _make_id(pfx: str = '', length=12) -> str:
    return pfx + ''.join(random.choice(string.ascii_letters) for _ in range(length))


def _prepare_proof_request(spec: ProofSpecCfg, wql_filters: dict = None) -> ProofRequest:
    """
    Prepare the JSON payload for a proof request

    Args:
        spec: the proof request specification
        wql_filters: a dict of WQL filters for the wallet
    """
    req_attrs = {}
    req_preds = {}
    for schema in spec.schemas:
        s_id = schema["definition"].schema_id
        s_uniq = hashlib.sha1(s_id.encode('ascii')).hexdigest()
        for attr in schema["attributes"]:
            req_attrs["{}_{}_uuid".format(s_uniq, attr)] = {
                "name": attr,
                "restrictions": [{
                    "schema_id": s_id,
                }]
            }
        for pred in schema.get("predicates") or []:
            req_preds["{}_{}_uuid".format(s_uniq, pred["name"])] = {
                "name": pred["name"],
                "p_type": pred["p_type"],
                "p_value": pred["p_value"],
                "restrictions": [{
                    "schema_id": s_id,
                }]
            }
    return ProofRequest({
        "name": spec.spec_id,
        "nonce": str(random.randint(10000000000, 100000000000)),  # FIXME - how best to generate?
        "version": spec.version,
        "requested_attributes": req_attrs,
        "requested_predicates": req_preds,
    }, wql_filters)


def _populate_cred_def_ids(proof_req: dict, creds: list):
    """
    Populate cred_def_id for each attribute in proof request if not defined
    """
    cdef_map = {}
    default_cred_def_id = None
    for cred in creds:
        # accept list of cred info or cred briefs
        if "cred_info" in cred:
            cred = cred["cred_info"]
        cdef_map[cred["schema_id"]] = cred["cred_def_id"]
        if len(creds) == 1:
            default_cred_def_id = cred["cred_def_id"]
    for attr in proof_req["requested_attributes"].values():
        attr_cdef_id = None
        attr_schema_id = None
        for rest in attr["restrictions"]:
            attr_cdef_id = rest.get("cred_def_id", attr_cdef_id)
            attr_schema_id = rest.get("schema_id", attr_schema_id)
        if not attr_cdef_id:
            if attr_schema_id and attr_schema_id in cdef_map:
                attr["restrictions"].append({
                    "cred_def_id": cdef_map[attr_schema_id],
                })
            elif default_cred_def_id:
                attr["restrictions"].append({
                    "cred_def_id": default_cred_def_id,
                })
        if "non_revoked" not in attr:
            attr["non_revoked"] = {}


class IndyService(ServiceBase):
    """
    A class for managing interactions with the Hyperledger Indy ledger
    """

    def __init__(self, pid: str, exchange: Exchange, env: Mapping, spec: dict = None):
        super(IndyService, self).__init__(pid, exchange, env)
        self._config = {}
        self._genesis_path = None
        self._agents = {}
        self._connections = {}
        self._ledger_url = None
        self._genesis_url = None
        self._protocol_version = None
        self._max_concurrent_storage = env.get("MAX_CONCURRENT_STORAGE", 20)
        self._name = pid
        self._opened = False
        self._pool = None
        self._proof_specs = {}
        self._storage_lock = None
        self._wallets = {}
        self._verifier = None
        self._update_config(spec)

    def _update_config(self, spec) -> None:
        """
        Load configuration settings
        """
        if spec:
            self._config.update(spec)
        if "name" in spec:
            self._name = spec["name"]
        if "ledger_url" in spec:
            self._ledger_url = spec["ledger_url"]
        if "genesis_url" in spec:
            self._genesis_url = spec["genesis_url"]
        if "protocol_version" in spec:
            self._protocol_version = spec["protocol_version"]

    async def _service_start(self) -> bool:
        """
        Initial service startup sequence
        """
        self._storage_lock = asyncio.Semaphore(self._max_concurrent_storage)
        LOGGER.info("Max concurrent: %s", self._max_concurrent_storage)
        return await super(IndyService, self)._service_start()

    async def _service_sync(self) -> bool:
        """
        Perform the initial setup of the ledger connection, including downloading the
        genesis transaction file
        """
        await self._setup_pool()
        synced = True
        for wallet in self._wallets.values():
            if not wallet.created:
                await wallet.create()
        for agent in self._agents.values():
            if not await self._sync_agent(agent):
                LOGGER.debug("Agent not yet synced: %s", agent.agent_id)
                synced = False
        for connection in self._connections.values():
            if not await self._sync_connection(connection):
                LOGGER.debug("Connection not yet synced: %s", connection.connection_id)
                synced = False
        for spec in self._proof_specs.values():
            if not await self._sync_proof_spec(spec):
                LOGGER.debug("Proof spec not synced: %s", spec.spec_id)
                synced = False
        return synced

    async def _service_stop(self) -> None:
        """
        Shut down active connections
        """
        for connection in self._connections.values():
            await connection.close()
        for agent in self._agents.values():
            await agent.close()
        for wallet in self._wallets.values():
            await wallet.close()

    def _add_agent(self, agent_type: str, wallet_id: str, **params) -> str:
        """
        Add an agent configuration

        Args:
            agent_type: the agent type, issuer or holder
            wallet_id: the identifier for a previously-registered wallet
            params: parameters to be passed to the :class:`AgentCfg` constructor
        """
        if wallet_id not in self._wallets:
            raise IndyConfigError("Wallet ID not registered: {}".format(wallet_id))
        cfg = AgentCfg(agent_type, wallet_id, **params)
        if not cfg.agent_id:
            cfg.agent_id = _make_id("agent-")
        if cfg.agent_id in self._agents:
            raise IndyConfigError("Duplicate agent ID: {}".format(cfg.agent_id))
        agents = self._agents.copy()
        agents[cfg.agent_id] = cfg
        self._agents = agents
        return cfg.agent_id

    def _get_agent_status(self, agent_id: str) -> ServiceResponse:
        """
        Return the status of a registered agent

        Args:
            agent_id: the unique identifier of the agent
        """
        if agent_id in self._agents:
            msg = AgentStatus(agent_id, self._agents[agent_id].status)
        else:
            msg = IndyServiceFail("Unregistered agent: {}".format(agent_id))
        return msg

    def _add_credential_type(self, issuer_id: str, schema_name: str,
                             schema_version: str, origin_did: str,
                             attr_names: Sequence, config: Mapping = None) -> None:
        """
        Add a credential type to a given issuer

        Args:
            issuer_id: the identifier of the issuer service
            schema_name: the name of the schema used by the credential type
            schema_version: the version of the schema used by the credential type
            origin_did: the DID of the service issuing the schema (optional)
            attr_names: a list of schema attribute names
            config: additional configuration for the credential type
        """
        agent = self._agents[issuer_id]
        if not agent:
            raise IndyConfigError("Agent ID not registered: {}".format(issuer_id))
        schema = SchemaCfg(schema_name, schema_version, attr_names, origin_did)
        agent.add_credential_type(schema, **(config or {}))

    def _add_connection(self, connection_type: str, agent_id: str, **params) -> str:
        """
        Add a connection configuration

        Args:
            connection_type: the type of the connection, normally TheOrgBook
            agent_id: the identifier of the registered agent
            params: parameters to be passed to the :class:`ConnectionCfg` constructor
        """
        if agent_id not in self._agents:
            raise IndyConfigError("Agent ID not registered: {}".format(agent_id))
        cfg = ConnectionCfg(
            connection_type, agent_id, self._agents[agent_id].agent_type.value, **params)
        if not cfg.connection_id:
            cfg.connection_id = _make_id("connection-")
        if cfg.connection_id in self._connections:
            raise IndyConfigError("Duplicate connection ID: {}".format(cfg.connection_id))
        conns = self._connections.copy()
        conns[cfg.connection_id] = cfg
        self._connections = conns
        return cfg.connection_id

    def _get_connection_status(self, connection_id: str) -> ServiceResponse:
        """
        Return the status of a registered connection

        Args:
            connection_id: the unique identifier of the connection
        """
        if connection_id in self._connections:
            msg = ConnectionStatus(connection_id, self._connections[connection_id].status)
        else:
            msg = IndyServiceFail("Unregistered connection: {}".format(connection_id))
        return msg

    def _add_wallet(self, **params) -> str:
        """
        Add a wallet configuration

        Args:
            params: parameters to be passed to the :class:`WalletCfg` constructor
        """
        cfg = WalletCfg(**params)
        if not cfg.wallet_id:
            cfg.wallet_id = _make_id("wallet-")
        if cfg.wallet_id in self._wallets:
            raise IndyConfigError("Duplicate wallet ID: {}".format(cfg.wallet_id))
        wallets = self._wallets.copy()
        wallets[cfg.wallet_id] = cfg
        self._wallets = wallets
        return cfg.wallet_id

    def _get_wallet_status(self, wallet_id: str) -> ServiceResponse:
        """
        Return the status of a registered wallet

        Args:
            wallet_id: the unique identifier of the wallet
        """
        if wallet_id in self._wallets:
            msg = WalletStatus(wallet_id, self._wallets[wallet_id].status)
        else:
            msg = IndyServiceFail("Unregistered wallet: {}".format(wallet_id))
        return msg

    async def _sync_agent(self, agent: AgentCfg) -> bool:
        """
        Perform agent synchronization, registering the DID and publishing schemas
        and credential definitions as required

        Args:
            agent: the Indy agent configuration
        """
        LOGGER.debug('Checking if agent synced...')
        if not agent.synced:
            LOGGER.debug('Syncing agent...')
            if not agent.created:
                LOGGER.debug('Creating agent...')
                wallet = self._wallets[agent.wallet_id]
                if not wallet.created:
                    return False
                await agent.create(wallet, self._pool)

            LOGGER.debug('Opening agent...')
            await agent.open()

            LOGGER.debug('Checking if agent is registered...')
            if not agent.registered:
                # check DID is registered
                LOGGER.debug('Registering agent...')
                auto_register = self._config.get("auto_register", True)
                await self._check_registration(agent, auto_register, agent.role)

                # check endpoint is registered (if any)
                LOGGER.debug('Checking agent endpoint...')
                await self._check_endpoint(agent)
                agent.registered = True

            # publish schemas
            for cred_type in agent.cred_types:
                LOGGER.debug('Publishing agent schemas...')
                await self._publish_schema(agent, cred_type)

            agent.synced = True
            LOGGER.info("Indy agent synced: %s", agent.agent_id)
        return agent.synced

    async def _sync_connection(self, connection: ConnectionCfg) -> bool:
        """
        Perform synchronization on a connection object
        """
        agent = self._agents[connection.agent_id]

        if not connection.synced:
            if not connection.created:
                if not agent.synced:
                    return False
                agent_cfg = agent.get_connection_params(connection)
                if not agent_cfg:
                    agent_cfg = {}
                agent_cfg["config_root"] = self._env.get("CONFIG_ROOT")
                await connection.create(agent_cfg)

            try:
                if not connection.opened:
                    await connection.open(self)

                await connection.sync()
            except IndyConnectionError as e:
                raise ServiceSyncError("Error syncing connection {}: {}".format(
                    connection.connection_id, str(e))) from None
        return connection.synced

    async def _setup_pool(self) -> None:
        """
        Initialize the Indy NodePool, fetching the genesis transaction if necessary
        """
        if not self._opened:
            await self._check_genesis_path()
            if self._protocol_version:
                pool_cfg = {'protocol': self._protocol_version}
            else:
                pool_cfg = None
            self._pool = NodePool(self._name, self._genesis_path, pool_cfg)
            await self._pool.open()
            self._opened = True

    async def _check_genesis_path(self) -> None:
        """
        Make sure that the genesis path is defined, and download the transaction file if needed.
        """
        if not self._genesis_path:
            path = self._config.get("genesis_path")
            if not path:
                raise IndyConfigError("Missing genesis_path")
            genesis_path = pathlib.Path(path)
            if not genesis_path.exists():
                genesis_url = self._genesis_url
                if not genesis_url:
                    ledger_url = self._ledger_url
                    if not ledger_url:
                        raise IndyConfigError(
                            "Cannot retrieve genesis transaction without ledger_url or genesis_url"
                        )
                    genesis_url = "{}/genesis".format(ledger_url)
                parent_path = pathlib.Path(genesis_path.parent)
                if not parent_path.exists():
                    parent_path.mkdir(parents=True)
                await self._fetch_genesis_txn(genesis_url, genesis_path)
            elif genesis_path.is_dir():
                raise IndyConfigError("genesis_path must not point to a directory")
            self._genesis_path = path

    async def _fetch_genesis_txn(self, genesis_url: str, target_path: str) -> bool:
        """
        Download the genesis transaction file from the ledger server

        Args:
            genesis_url: the root address of genesis file
            target_path: the filesystem path of the genesis transaction file once downloaded
        """
        LOGGER.info(
            "Fetching genesis transaction file from %s", genesis_url
        )

        try:
            async with HttpSession('fetching genesis transaction', timeout=15) as handler:
                response = await handler.client.get(genesis_url)
                await handler.check_status(response, (200,))
                data = await response.text()
        except IndyConnectionError as e:
            raise ServiceSyncError(str(e)) from None

        # check data is valid json
        LOGGER.debug("Genesis transaction response: %s", data)
        lines = data.splitlines()
        if not lines or not json.loads(lines[0]):
            raise ServiceSyncError("Genesis transaction file is not valid JSON")

        # write result to provided path
        with target_path.open("x") as output_file:
            output_file.write(data)
        return True

    async def _check_registration(self, agent: AgentCfg, auto_register: bool = True,
                                  role: str = "") -> None:
        """
        Look up our nym on the ledger and register it if not present

        Args:
            agent: the initialized and opened agent to be checked
            auto_register: whether to automatically register the DID on the ledger
        """
        did = agent.did
        LOGGER.debug("Checking DID registration %s", did)
        nym_json = await agent.instance.get_nym(did)
        LOGGER.debug("get_nym result for %s: %s", did, nym_json)

        nym_info = json.loads(nym_json)
        if not nym_info:
            if not auto_register:
                raise ServiceSyncError(
                    "DID is not registered on the ledger and auto-registration disabled"
                )

            ledger_url = self._ledger_url
            if not ledger_url:
                raise IndyConfigError("Cannot register DID without ledger_url")
            LOGGER.info("Registering DID %s", did)

            try:
                async with HttpSession('DID registration', timeout=30) as handler:
                    response = await handler.client.post(
                        "{}/register".format(ledger_url),
                        json={"did": did, "verkey": agent.verkey, "role": role},
                    )
                    await handler.check_status(response, (200,))
                    nym_info = await response.json()
            except IndyConnectionError as e:
                raise ServiceSyncError(str(e)) from None
            LOGGER.debug("Registration response: %s", nym_info)
            if not nym_info or not nym_info["did"]:
                raise ServiceSyncError(
                    "DID registration failed: {}".format(nym_info)
                )

    async def _check_endpoint(self, agent: AgentCfg) -> None:
        """
        Look up our endpoint on the ledger and register it if not present

        Args:
            agent: the initialized and opened agent to be checked
            endpoint: the endpoint to be added to the ledger, if not defined
        """
        await agent.send_endpoint()
        LOGGER.info("Endpoint stored: %s", agent.endpoint)

    async def _publish_schema(self, issuer: AgentCfg, cred_type: dict) -> None:
        """
        Check the ledger for a specific schema and version, and publish it if not found.
        Also publish the related credential definition if not found

        Args:
            issuer: the initialized and opened issuer instance publishing the schema
            cred_type: a dict which will be updated with the published schema and credential def
        """

        if not cred_type or "definition" not in cred_type:
            raise IndyConfigError("Missing schema definition")
        definition = cred_type["definition"]
        s_id = schema_id(issuer.did, definition.name, definition.version)

        if not cred_type.get("ledger_schema"):
            LOGGER.info(
                "Checking for schema: %s (%s)",
                definition.name,
                definition.version,
            )
            # Check if schema exists on ledger

            try:
                s_key = schema_key(s_id)
                schema_json = await issuer.instance.get_schema(s_key)
                ledger_schema = json.loads(schema_json)
                log_json("Schema found on ledger:", ledger_schema, LOGGER)
                if sorted(ledger_schema["attrNames"]) != sorted(definition.attr_names):
                    raise IndyConfigError(
                        "Ledger schema attributes do not match definition, found: {}".format(
                            ledger_schema["attrNames"]))
            except AbsentSchema:
                # If not found, send the schema to the ledger
                LOGGER.info(
                    "Publishing schema: %s (%s)",
                    definition.name,
                    definition.version,
                )
                schema_json = await issuer.instance.send_schema(
                    json.dumps(
                        {
                            "name": definition.name,
                            "version": definition.version,
                            "attr_names": definition.attr_names,
                        }
                    )
                )
                ledger_schema = json.loads(schema_json)
                if not ledger_schema or not ledger_schema.get("seqNo"):
                    raise ServiceSyncError("Schema was not published to ledger")
                log_json("Published schema:", ledger_schema, LOGGER)
            cred_type["ledger_schema"] = ledger_schema

        if not cred_type.get("cred_def"):
            # Check if credential definition has been published
            LOGGER.info(
                "Checking for credential def: %s (%s)",
                definition.name,
                definition.version,
            )

            try:
                cred_def_json = await issuer.instance.get_cred_def(
                    cred_def_id(issuer.did, cred_type["ledger_schema"]["seqNo"], self._pool.protocol)
                )
                cred_def = json.loads(cred_def_json)
                log_json("Credential def found on ledger:", cred_def, LOGGER)
            except AbsentCredDef:
                # If credential definition is not found then publish it
                LOGGER.info(
                    "Publishing credential def: %s (%s)",
                    definition.name,
                    definition.version,
                )
                cred_def_json = await issuer.instance.send_cred_def(
                    s_id, revocation=False
                )
                cred_def = json.loads(cred_def_json)
                log_json("Published credential def:", cred_def, LOGGER)
            cred_type["cred_def"] = cred_def

    async def _issue_credential(self, connection_id: str, schema_name: str,
                                schema_version: str, origin_did: str,
                                cred_data: Mapping,
                                batch: bool = False) -> ServiceResponse:
        """
        Issue a credential to the connection target

        Args:
            connection_id: the identifier of the registered connection
            schema_name: the name of the credential schema
            schema_version: the version of the credential schema
            origin_did: the origin DID of the ledger schema (may be None)
            cred_data: the raw credential attributes
        """
        conn = self._connections.get(connection_id)
        if not conn:
            raise IndyConfigError("Unknown connection id: {}".format(connection_id))
        if not conn.synced:
            raise IndyConfigError("Connection is not yet synchronized: {}".format(connection_id))
        issuer = self._agents[conn.agent_id]
        if not issuer.is_issuer:
            raise IndyConfigError(
                "Cannot issue credential from non-issuer agent: {}".format(issuer.agent_id))
        if not issuer.synced:
            raise IndyConfigError("Issuer is not yet synchronized: {}".format(issuer.agent_id))
        cred_type = issuer.find_credential_type(schema_name, schema_version, origin_did)
        if not cred_type:
            raise IndyConfigError("Could not locate credential type: {}/{} {}".format(
                schema_name, schema_version, origin_did))

        cred_request_cache = cred_type.get("cred_request_cache")
        if not cred_request_cache:
            cred_request_cache = cred_type["cred_request_cache"] = \
                {"request": None, "lock": asyncio.Lock()}
        async with cred_request_cache["lock"]:
            if cred_request_cache["request"] and cred_request_cache["expiry"] > time.time():
                cred_request = cred_request_cache["request"]
                LOGGER.debug("Fetched credential request from cache")
            else:
                cred_offer = await self._create_cred_offer(issuer, cred_type)
                log_json("Created cred offer:", cred_offer, LOGGER)
                cred_request = await conn.instance.generate_credential_request(cred_offer)
                cred_request_cache["request"] = cred_request
                cred_request_cache["expiry"] = time.time() + 600
                LOGGER.debug("Saved cred request cache")
        log_json("Got cred request:", cred_request, LOGGER)

        async def make_cred(cred_data):
            fixed_data = self._fix_cred_data(cred_type["definition"], cred_data)
            cred = await self._create_cred(issuer, cred_request, fixed_data)
            log_json("Created cred:", cred, LOGGER)
            return cred

        if batch:
            creds = []
            for data in cred_data:
                creds.append(asyncio.ensure_future(make_cred(data)))
            creds = await asyncio.gather(*creds)
            stored = await conn.instance.store_credential_batch(creds)
            log_json("Stored credentials:", stored, LOGGER)
        else:
            cred = await make_cred(cred_data)
            stored = await conn.instance.store_credential(cred)
            log_json("Stored credential:", stored, LOGGER)

        return stored

    def _fix_cred_data(self, schema, cred_data: dict):
        """
        Provide empty values for any missing schema attributes and remove unknown
        attributes from the credential data

        Args:
            schema: the schema definition
            cred_data: the dictionary of schema attributes
        """
        ret = {}
        for key in schema.attr_names:
            ret[key] = cred_data.get(key)
        return ret

    async def _create_cred_offer(self, issuer: AgentCfg,
                                 cred_type) -> CredentialOffer:
        """
        Create a credential offer for a specific connection from a given issuer

        Args:
            issuer: the issuer configuration object
            cred_type: the credential type definition
        """
        schema = cred_type["definition"]

        LOGGER.info(
            "Creating Indy credential offer for issuer %s, schema %s",
            issuer.agent_id,
            schema.name,
        )
        cred_offer_json = await issuer.instance.create_cred_offer(
            cred_type["ledger_schema"]["seqNo"]
        )
        return CredentialOffer(
            json.loads(cred_offer_json),
            cred_type["cred_def"]["id"],
        )

    async def _create_cred(self, issuer: AgentCfg, request: CredentialRequest,
                           cred_data: Mapping) -> Credential:
        """
        Create a credential from a credential request for a specific issuer

        Args:
            issuer: the issuer configuration object
            request: a credential request returned from the holder service
            cred_data: the raw credential attributes
        """
        async with self._storage_lock:
            (cred_json, cred_revoc_id, _epoch_creation) = await issuer.instance.create_cred(
                json.dumps(request.cred_offer.data),
                request.data,
                cred_data,
            )
        return Credential(
            json.loads(cred_json),
            request.metadata,
            cred_revoc_id,
        )

    async def _generate_credential_request(
            self, holder_id: str, cred_offer: CredentialOffer) -> CredentialRequest:
        """
        Generate a credential request for a given holder agent from a credential offer
        """
        holder = self._agents.get(holder_id)
        if not holder:
            raise IndyConfigError("Unknown holder id: {}".format(holder_id))
        if not holder.is_holder:
            raise IndyConfigError(
                "Cannot generate credential request from non-holder agent: {}".format(holder.agent_id))
        async with self._storage_lock:
            if not holder.synced:
                raise IndyConfigError("Holder is not yet synchronized: {}".format(holder_id))
            (cred_req, req_metadata_json) = await holder.instance.create_cred_req(
                json.dumps(cred_offer.data),
                cred_offer.cred_def_id,
            )
        return CredentialRequest(
            cred_offer,
            cred_req,
            json.loads(req_metadata_json),
        )

    async def _store_credential(self, holder_id: str,
                                credential: Credential) -> StoredCredential:
        """
        Store a credential in a given holder agent's wallet
        """
        holder = self._agents.get(holder_id)
        if not holder:
            raise IndyConfigError("Unknown holder id: {}".format(holder_id))
        if not holder.is_holder:
            raise IndyConfigError(
                "Cannot store credential using non-holder agent: {}".format(holder.agent_id))
        async with self._storage_lock:
            if not holder.synced:
                raise IndyConfigError("Holder is not yet synchronized: {}".format(holder_id))
            cred_id = await holder.instance.store_cred(
                json.dumps(credential.cred_data),
                json.dumps(credential.cred_req_metadata),
            )
        return StoredCredential(
            credential,
            cred_id,
        )

    async def _resolve_schema(self, schema_name: str, schema_version: str,
                              origin_did: str) -> ResolvedSchema:
        """
        Resolve a schema defined by one of our issuers
        """
        lookup_agent = None
        for agent_id, agent in self._agents.items():
            if agent.synced:
                found = agent.find_credential_type(schema_name, schema_version, origin_did)
                if found:
                    defn = found["definition"]
                    did = defn.origin_did or agent.did
                    return ResolvedSchema(
                        agent_id,
                        schema_id(did, defn.name, defn.version),
                        defn.name,
                        defn.version,
                        did,
                        defn.attr_names,
                    )
                lookup_agent = agent
        if schema_name and schema_version and origin_did and lookup_agent:
            s_id = schema_id(origin_did, schema_name, schema_version)
            s_key = schema_key(s_id)
            try:
                schema_json = await lookup_agent.instance.get_schema(s_key)
                ledger_schema = json.loads(schema_json)
                log_json("Schema found on ledger:", ledger_schema, LOGGER)
                return ResolvedSchema(
                    None,
                    s_id,
                    schema_name,
                    schema_version,
                    origin_did,
                    ledger_schema["attrNames"],
                )
            except AbsentSchema:
                pass
        raise IndyConfigError("Issuer schema not found: {}/{}".format(schema_name, schema_version))

    async def _construct_proof(self, holder_id: str, proof_req: ProofRequest,
                               cred_ids: set = None) -> ConstructedProof:
        """
        Construct a proof from credentials in the holder's wallet, given a proof request
        """
        holder = self._agents.get(holder_id)
        if not holder:
            raise IndyConfigError("Unknown holder id: {}".format(holder_id))
        if not holder.is_holder:
            raise IndyConfigError(
                "Cannot construct proof from non-holder agent: {}".format(holder.agent_id))
        if not holder.synced:
            raise IndyConfigError("Holder is not yet synchronized: {}".format(holder_id))
        log_json("Fetching credentials for request", proof_req.data, LOGGER)

        # TODO - use separate request to find credentials and allow manual filtering?
        if cred_ids:
            LOGGER.debug("Construct proof from IDs: %s", cred_ids)
            found_creds = []
            for cred_id in cred_ids:
                try:
                    found_cred_json = await holder.instance.get_cred_info_by_id(
                        cred_id,
                    )
                    found_creds.append(json.loads(found_cred_json))
                except AbsentCred:
                    LOGGER.warning("Credential not found: %s", cred_id)

            if not found_creds:
                raise IndyError("No credentials found for proof")
            _populate_cred_def_ids(proof_req.data, found_creds)
            found_creds = proof_req_infos2briefs(proof_req.data, found_creds)
        else:
            # DEBUG
            #proof_req.wql_filters = {
            #    'eb8cce736d877a9a45bcc62303e62d25c0fe5da6_attr1_uuid': {
            #        'attr::attr1::value': '5'
            #    }
            #}

            _cred_ids, found_creds_json = await holder.instance.get_cred_briefs_by_proof_req_q(
                json.dumps(proof_req.data),
                json.dumps(proof_req.wql_filters) if proof_req.wql_filters else None,
            )
            found_creds = json.loads(found_creds_json)
            _populate_cred_def_ids(proof_req.data, found_creds)

        log_json("Found credentials", found_creds, LOGGER)

        if not found_creds:
            raise IndyError("No credentials found for proof")
        elif len(found_creds) > 1:
            raise IndyError("Too many credentials found for proof")

        request_params = proof_req_briefs2req_creds(proof_req.data, found_creds)

        # FIXME catch exception?
        log_json("Creating proof", request_params, LOGGER)
        proof_json = await holder.instance.create_proof(
            proof_req.data,
            found_creds,
            request_params,
        )
        proof = json.loads(proof_json)
        return ConstructedProof(proof)

    def _add_proof_spec(self, **params) -> str:
        """
        Add a proof request specification

        Args:
            params: parameters to be passed to the :class:`ProofSpecCfg` constructor
        """
        cfg = ProofSpecCfg(**params)
        if not cfg.spec_id:
            cfg.spec_id = _make_id("proof-")
        if cfg.spec_id in self._proof_specs:
            raise IndyConfigError("Duplicate proof spec ID: {}".format(cfg.spec_id))
        self._proof_specs[cfg.spec_id] = cfg
        return cfg.spec_id

    async def _sync_proof_spec(self, spec: ProofSpecCfg) -> bool:
        """
        Resolve schema information for a proof specification
        """
        missing = spec.get_incomplete_schemas()
        check = False
        for s_key in missing:
            try:
                found = await self._resolve_schema(*s_key)
                cfg = SchemaCfg(
                    found.schema_name, found.schema_version,
                    found.attr_names, found.origin_did)
                spec.populate_schema(cfg)
                check = True
            except IndyConfigError:
                pass
        if check:
            missing = spec.get_incomplete_schemas()
        spec.synced = not missing
        return spec.synced

    def _get_proof_spec_status(self, spec_id: str) -> ServiceResponse:
        """
        Return the status of a registered proof spec

        Args:
            spec_id: the unique identifier of the proof specification
        """
        if spec_id in self._proof_specs:
            msg = ProofSpecStatus(spec_id, self._proof_specs[spec_id].status)
        else:
            msg = IndyServiceFail("Unregistered proof spec: {}".format(spec_id))
        return msg

    async def _generate_proof_request(self, spec_id: str, wql_filters: dict = None) -> ProofRequest:
        """
        Create a proof request from a previously registered proof specification
        """
        spec = self._proof_specs.get(spec_id)
        if not spec:
            raise IndyConfigError("Proof specification not defined: {}".format(spec_id))
        if not spec.synced:
            raise IndyConfigError("Proof specification not synced: {}".format(spec_id))
        return _prepare_proof_request(spec, wql_filters)

    async def _request_proof(self, connection_id: str, proof_req: ProofRequest,
                             cred_ids: set = None, params: dict = None) -> VerifiedProof:
        """
        Request a verified proof from a connection
        """
        conn = self._connections.get(connection_id)
        if not conn:
            raise IndyConfigError("Unknown connection id: {}".format(connection_id))
        if not conn.synced:
            raise IndyConfigError("Connection is not yet synchronized: {}".format(connection_id))
        verifier = self._agents[conn.agent_id]
        if not verifier.is_verifier:
            raise IndyConfigError(
                "Cannot verify proof from non-verifier agent: {}".format(verifier.agent_id))
        if not verifier.synced:
            raise IndyConfigError("Verifier is not yet synchronized: {}".format(verifier.agent_id))
        proof = await conn.instance.construct_proof(proof_req, cred_ids, params)
        return await self._verify_proof(verifier.agent_id, proof_req, proof)

    async def _verify_proof(self, verifier_id: str, proof_req: ProofRequest,
                            proof: ConstructedProof) -> VerifiedProof:
        """
        Verify a constructed proof

        Args:
            verifier_id: the verifier agent to employ
            proof_req: the proof request to verify against
            proof: the constructed proof to verify
        """
        verifier = self._agents.get(verifier_id)
        if not verifier:
            raise IndyConfigError("Unknown verifier id: {}".format(verifier_id))
        if not verifier.synced:
            raise IndyConfigError("Verifier is not yet synchronized: {}".format(verifier.agent_id))
        result = await verifier.instance.verify_proof(proof_req.data, proof.proof)
        parsed_proof = revealed_attrs(proof.proof)
        return VerifiedProof(result, parsed_proof, proof)

    async def _resolve_nym(self, did: str, agent_id: str = None) -> ResolvedNym:
        """
        Resolve a DID on the ledger

        Args:
            did: the DID to resolve
            agent_id: the agent instance to employ
        """
        if not agent_id:
            for aid in self._agents:
                if self._agents[aid].synced:
                    agent_id = aid
                    break
        agent = self._agents.get(agent_id)
        if not agent:
            raise IndyConfigError("Unknown agent id: {}".format(agent_id))
        if not agent.synced:
            raise IndyConfigError("Agent is not yet synchronized: {}".format(agent.agent_id))
        nym_json = await agent.instance.get_nym(did)
        nym_info = json.loads(nym_json)
        if not nym_info:
            nym_info = None
        return ResolvedNym(did, nym_info)

    async def _handle_ledger_status(self):
        """
        Download the ledger status from von-network and return it to the client
        """
        url = self._ledger_url
        async with self.http as client:
            response = await client.get("{}/status".format(url))
        return await response.text()

    def _connection_http_client(self, conn_id: str = None, **kwargs):
        """
        Create a new :class:`ClientSession` which includes DID signing information in each request

        Args:
            conn_id: an optional identifier for a specific connection (to enable DID signing)
        Returns:
            the initialized :class:`ClientSession` object
        """
        if "request_class" not in kwargs:
            kwargs["request_class"] = SignedRequest
        if conn_id and "auth" not in kwargs:
            kwargs["auth"] = self._signed_request_auth(conn_id)
        return super(IndyService, self).http_client(**kwargs)

    def _signed_request_auth(self, conn_id: str, header_list=None):
        """
        Create a :class:`SignedRequestAuth` representing our authentication credentials,
        used to sign outgoing requests

        Args:
            conn_id: the unique identifier of the connection
            header_list: optionally override the list of headers to sign
        """
        conn = self._connections.get(conn_id)
        if not conn:
            raise IndyConfigError("Unknown connection ID: {}".format(conn_id))
        agent = self._agents[conn.agent_id]
        wallet = self._wallets[agent.wallet_id]
        if agent.did and wallet.seed:
            key_id = "did:sov:{}".format(agent.did)
            secret = wallet.seed
            if isinstance(secret, str):
                if secret[-1:] == "=":
                    secret = base64.b64decode(secret)
                else:
                    secret = bytes(secret, "ascii")
            ret = SignedRequestAuth(key_id, "ed25519", secret, header_list)
            if not conn.sign_target and hasattr(ret, "sign_target"):
                ret.sign_target = False
            return ret
        return None

    async def _service_request(self, request: ServiceRequest) -> ServiceResponse:
        """
        Process a message from the exchange and send the reply, if any

        Args:
            request: the message to be processed
        """
        if isinstance(request, LedgerStatusReq):
            with self._timer("ledger_status"):
                text = await self._handle_ledger_status()
                reply = LedgerStatus(text)

        elif isinstance(request, RegisterAgentReq):
            try:
                agent_id = self._add_agent(request.agent_type, request.wallet_id, **request.config)
                reply = self._get_agent_status(agent_id)
                self._sync_required()
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, RegisterConnectionReq):
            try:
                connection_id = self._add_connection(
                    request.connection_type, request.agent_id, **request.config)
                reply = self._get_connection_status(connection_id)
                self._sync_required()
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, RegisterCredentialTypeReq):
            try:
                self._add_credential_type(
                    request.issuer_id,
                    request.schema_name,
                    request.schema_version,
                    request.origin_did,
                    request.attr_names,
                    request.config)
                reply = IndyServiceAck()
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, RegisterWalletReq):
            try:
                wallet_id = self._add_wallet(**request.config)
                reply = self._get_wallet_status(wallet_id)
                self._sync_required()
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, AgentStatusReq):
            reply = self._get_agent_status(request.agent_id)

        elif isinstance(request, ConnectionStatusReq):
            reply = self._get_connection_status(request.connection_id)

        elif isinstance(request, WalletStatusReq):
            reply = self._get_wallet_status(request.wallet_id)

        elif isinstance(request, IssueCredentialReq):
            try:
                with self._timer("issue_credential"):
                    reply = await self._issue_credential(
                        request.connection_id,
                        request.schema_name,
                        request.schema_version,
                        request.origin_did,
                        request.cred_data)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, IssueCredentialBatchReq):
            try:
                with self._timer("issue_credential"):
                    reply = await self._issue_credential(
                        request.connection_id,
                        request.schema_name,
                        request.schema_version,
                        request.origin_did,
                        request.cred_data,
                        True)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, GenerateCredentialRequestReq):
            try:
                with self._timer("generate_credential_request"):
                    reply = await self._generate_credential_request(
                        request.holder_id, request.cred_offer)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, StoreCredentialReq):
            try:
                with self._timer("store_credential"):
                    reply = await self._store_credential(
                        request.holder_id, request.credential)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, ResolveSchemaReq):
            try:
                with self._timer("resolve_schema"):
                    reply = await self._resolve_schema(
                        request.schema_name, request.schema_version, request.origin_did)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, ConstructProofReq):
            try:
                with self._timer("construct_proof"):
                    reply = await self._construct_proof(
                        request.holder_id, request.proof_req, request.cred_ids)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, RegisterProofSpecReq):
            try:
                spec_id = self._add_proof_spec(**request.config)
                reply = self._get_proof_spec_status(spec_id)
                self._sync_required()
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, GenerateProofRequestReq):
            try:
                reply = await self._generate_proof_request(request.spec_id, request.wql_filters)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, RequestProofReq):
            try:
                with self._timer("request_proof"):
                    reply = await self._request_proof(
                        request.connection_id, request.proof_req,
                        request.cred_ids, request.params)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, VerifyProofReq):
            try:
                with self._timer("verify_proof"):
                    reply = await self._verify_proof(
                        request.verifier_id, request.proof_req, request.proof)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        elif isinstance(request, ResolveNymReq):
            try:
                with self._timer("resolve_nym"):
                    reply = await self._resolve_nym(request.did, request.agent_id)
            except IndyError as e:
                reply = IndyServiceFail(str(e))

        else:
            reply = None
        return reply



#!/usr/bin/env python3
#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

#
# "requests" must be installed - pip3 install requests
#

import argparse
import asyncio
import json
import os
import sys

import aiohttp

DEFAULT_AGENT_URL = os.environ.get('AGENT_URL', 'http://localhost:5000')

parser = argparse.ArgumentParser(
    description='Issue one or more credentials via von-x')
parser.add_argument('conn_id', help='the connection ID')
parser.add_argument('name', help='the proof request ID')
parser.add_argument('source_ids', nargs='+',
    help='the source IDs to use in the proof requests')
parser.add_argument('-u', '--url', default=DEFAULT_AGENT_URL,
    help='the URL of the von-x service')

args = parser.parse_args()

AGENT_URL = args.url
CONN_ID = args.conn_id
ENTITY_IDS = args.source_ids
PROOF_NAME = args.name

async def request_proof(http_client, conn_id, proof_name, proof_params):
    print('Requesting proof: {} {}'.format(proof_name, proof_params))

    try:
        response = await http_client.post(
            '{}/request-proof'.format(AGENT_URL),
            params={'connection_id': conn_id, 'name': proof_name},
            json={'params': proof_params},
        )
        if response.status != 200:
            raise RuntimeError(
                'Proof request could not be processed: {}'.format(await response.text())
            )
        result_json = await response.json()
    except Exception as exc:
        raise Exception(
            'Could not complete proof request. '
            'Are von-x and TheOrgBook running?') from exc

    print('Response from von-x:\n\n{}\n'.format(result_json))

async def request_all(conn_id, proof_name, entity_ids):
    async with aiohttp.ClientSession() as http_client:
        for entity_id in entity_ids:
            await request_proof(http_client, conn_id, proof_name, {'source_id': entity_id})

asyncio.get_event_loop().run_until_complete(request_all(CONN_ID, PROOF_NAME, ENTITY_IDS))



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implemention of the generic :class:`ServiceManager` class which is used to manage
a collection of :class:`ServiceBase` instances
"""


import logging
import os
from typing import Mapping

from . import config
from . import exchange as exch
from .service import (
    ServiceBase,
    ServiceStatus,
    ServiceStatusReq,
    ServiceResponse)

LOGGER = logging.getLogger(__name__)


class ServiceManager(ServiceBase):
    """
    The standard :class:`ServiceManager` class is responsible for starting the
    message exchange, registering itself as a service, starting any dependent
    services, and checking the status of those services. It should normally be run
    with `start_process()` before the web server process has forked.
    """

    def __init__(self, env: Mapping = None, pid: str = "manager"):
        super(ServiceManager, self).__init__(pid, exch.Exchange(), env or {})
        self._executor_cls = exch.RequestExecutor
        self._proc_locals = {"pid": os.getpid()}
        self._services = {}
        self._init_services()

    def _init_services(self) -> None:
        """
        Initialize all dependent services
        """
        pass

    def add_service(self, svc_id: str, service: ServiceBase):
        """
        Add a service to the service manager instance

        Args:
            svc_id: the unique identifier for the service
            service: the service instance
        """
        self._services[svc_id] = service

    async def get_service_status(self, svc_id: str) -> dict:
        """
        Fetch the status of a registered service

        Args:
            svc_id: the unique identifier for the service
        """
        pid = self.get_service(svc_id).pid
        result = await self.executor.submit(pid, ServiceStatusReq())
        if isinstance(result, ServiceStatus):
            return result.status
        else:
            raise RuntimeError("Unexpected response to status request: {}".format(result))

    def start(self, wait: bool = True) -> None:
        """
        Start the message processor and any other services
        """
        self._exchange.start(False)
        super(ServiceManager, self).start(wait)

    async def _service_start(self) -> bool:
        """
        Start all registered services
        """
        for _svc_id, service in self._services.items():
            service.start(True)
        return True

    def stop(self, wait: bool = True) -> None:
        """
        Stop the message processor and any other services
        """
        super(ServiceManager, self).stop(wait)
        self._exchange.stop()

    async def _service_stop(self) -> None:
        """
        Stop all registered services
        """
        LOGGER.debug("Stopping managed services")
        for _id, service in self._services.items():
            service.stop()

    async def _get_status(self) -> ServiceResponse:
        """
        Return the current status of the service
        """
        status = self._status.copy()
        status["services"] = {}
        for svc_id in self._services:
            status["services"][svc_id] = await self.get_service_status(svc_id)
        return ServiceStatus(status)

    @property
    def env(self) -> dict:
        """
        Accessor for our local environment dict
        """
        return self._env

    @property
    def exchange(self) -> exch.Exchange:
        """
        Accessor for the Exchange this ServiceManager uses for messaging
        """
        return self._exchange

    @property
    def proc_locals(self) -> dict:
        """
        Accessor for all process-local variables

        Returns:
            a dictionary of currently-defined variables
        """
        pid = os.getpid()
        if self._proc_locals["pid"] != pid:
            self._proc_locals = {"pid": pid}
        return self._proc_locals

    @property
    def executor(self) -> exch.RequestExecutor:
        """
        Return a per-process request executor which manages requests
        and polls for results coming from other services.
        Note: this is called for each worker process started by the webserver.
        """
        ploc = self.proc_locals
        if not "executor" in ploc:
            ident = "exec-{}".format(ploc["pid"])
            ploc["executor"] = self._executor_cls(ident, self._exchange)
            ploc["executor"].start()
        return ploc["executor"]

    def get_service(self, name: str):
        """
        Fetch a defined service by name

        Args:
            name: the string identifier for the service

        Returns:
            the service instance, or None if not found
        """
        if name == "manager":
            return self
        return self._services.get(name)

    def get_service_message_target(self, name: str) -> exch.MessageTarget:
        """
        Get an endpoint for one of the services defined by this manager.
        This Endpoint can be used for sending process-safe messages and receiving results.

        Args:
            name: the string identifier for the service
            loop: the current event loop, if any
        """
        svc = self.get_service(name)
        if svc:
            return self.executor.get_message_target(svc.pid)
        return None

    def get_service_request_target(self, name: str) -> exch.RequestTarget:
        """
        Get an endpoint for sending messages to a service on the message exchange.
        Requests will be handled by the executor for this manager in this process.

        Args:
            name: the string identifier for the service
            loop: the current event loop, if any
        """
        ploc = self.proc_locals
        tg_name = "target_" + name
        if tg_name not in ploc:
            svc = self.get_service(name)
            if svc:
                ploc[tg_name] = self.executor.get_request_target(svc.pid)
            else:
                return None
        return ploc[tg_name]


class ConfigServiceManager(ServiceManager):
    """
    A :class:`ServiceManager` subclass with standard configuration loading methods
    """

    def __init__(self, env: Mapping = None, pid: str = "manager"):
        super(ConfigServiceManager, self).__init__(env, pid)
        self._services_cfg = None

    @property
    def config_root(self) -> str:
        """
        Accessor for the value of the CONFIG_ROOT setting, defaulting to the current directory
        """
        return self._env.get("CONFIG_ROOT") or os.curdir

    def load_config_path(self, settings_key, default_path, env=None) -> dict:
        """
        Load a YAML configuration file with defined variables replaced in the result

        Args:
            settings_key: the name of an environment variable defining an alternative
                configuration path
            default_path: the default path to the configuration file

        Returns:
            the parsed YAML configuration with variables replaced
        """
        path = self._env.get(settings_key)
        if not path:
            path = os.path.join(self.config_root, default_path)
        return config.load_config(path, env or self._env)

    def services_config(self, section: str) -> dict:
        """
        Load a named section from the global services.yml configuration

        Args:
            section: the configuration key
        """
        if self._services_cfg is None:
            self._services_cfg = self.load_config_path("SERVICES_CONFIG_PATH", "services.yml")
        if self._services_cfg:
            return self._services_cfg.get(section) or {}
        return {}



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Utility functions and classes
"""

import json
import logging
import time

from .exchange import ExchangeMessage


class MessageEncoder(json.JSONEncoder):
    """
    Customize JSONEncoder to automatically encode :class:`ExchangeMessage` instances
    """
    def default(self, o):
        if isinstance(o, ExchangeMessage):
            return dict(o)
        return super(MessageEncoder, self).default(o)


class JsonRepr:
    """
    Utility class to avoid JSON encoding debug output unless needed
    """
    def __init__(self, value, indent=2):
        self.value = value
        self.indent = indent

    def __repr__(self):
        return json.dumps(self.value, indent=self.indent, cls=MessageEncoder)


def log_json(heading, data, logger=None, level=logging.DEBUG):
    """
    Utility method to log JSON data for debugging
    """
    if not logger:
        logger = logging.getLogger(__name__)
    msg = """
============================================================================
%s
%s
============================================================================
"""
    logger.log(level, msg, heading, JsonRepr(data))


def normalize_credential_ids(cred_ids) -> set:
    """
    Clean up credential ID input
    """
    if isinstance(cred_ids, str):
        cred_ids = [id.strip() for id in cred_ids.split(",")]
    if isinstance(cred_ids, list):
        cred_ids = set(filter(None, cred_ids))
    elif not isinstance(cred_ids, set):
        cred_ids = None
    return cred_ids



class Stats:
    """
    Measure combined statistics for various named tasks
    """

    class Timer:
        """
        An instance of a timer that can be used in a with statement
        """
        def __init__(self, stats, tasks, log_as=None):
            self.duration = None
            self.handle = None
            self.log_as = log_as
            self.stats = stats
            self.tasks = tasks

        def start(self):
            """
            Start the timer
            """
            self.handle = self.stats.start(*self.tasks, log_as=self.log_as)
            return self

        def end(self):
            """
            End the timer
            """
            if self.duration is None:
                self.duration = self.stats.end(self.handle)
            return self.duration

        def __enter__(self):
            return self.start()

        def __exit__(self, exception_type, exception_value, traceback):
            self.end()

    def __init__(self, logger=None, log_level=logging.DEBUG):
        self.count = {}
        self.current = {}
        self.logger = logger
        self.log_level = log_level
        self.max = {}
        self.min = {}
        self.total = {}

    def start(self, *tasks, log_as=None):
        """
        Start a new set of tasks
        """
        if tasks and not log_as:
            log_as = tasks[0]
        if log_as and self.logger:
            self.logger.log(self.log_level, ">>> %s", log_as)
        for task in tasks:
            self.current[task] = self.current.get(task, 0) + 1
        return (time.perf_counter(), tasks, log_as)

    def end(self, handle):
        """
        End a previously started set of tasks
        """
        (start, tasks, log_as) = handle
        diff = time.perf_counter() - start
        for task in tasks:
            self.current[task] -= 1
            if task in self.count:
                self.count[task] += 1
                self.max[task] = max(self.max[task], diff)
                self.min[task] = min(self.min[task], diff)
                self.total[task] += diff
            else:
                self.count[task] = 1
                self.max[task] = diff
                self.min[task] = diff
                self.total[task] = diff
        if log_as and self.logger:
            self.logger.log(self.log_level, "<<< %s (%0.5f)", log_as, diff)
        return diff

    def timer(self, *tasks, log_as=None):
        """
        Create a new timer for a set of tasks
        """
        return self.Timer(self, tasks, log_as=log_as)

    def results(self):
        return {
            "avg": {task: self.total[task] / self.count[task] for task in self.count},
            "count": self.count.copy(),
            "current": self.current.copy(),
            "max": self.max.copy(),
            "min": self.min.copy(),
            "total": self.total.copy(),
        }



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Classes for managing the active :class:`IndyService` configuration - agents, connections,
schemas, proof requests, and wallets.
"""

import base64
import binascii
from distutils.version import LooseVersion
from enum import Enum
import logging
from typing import Mapping, Sequence
import uuid

from von_anchor import (
    Issuer,
    HolderProver,
    Verifier,
)
from von_anchor.anchor.base import _BaseAnchor
from von_anchor.anchor.demo import BCRegistrarAnchor, OrgHubAnchor
from von_anchor.nodepool import NodePool
from von_anchor.wallet import Wallet, register_wallet_storage_library
from von_anchor.util import schema_id

from indy.error import IndyError, ErrorCode

from .connection import ConnectionBase, ConnectionType, HolderConnection, HttpConnection
from .errors import IndyConfigError
from .tob import CRED_TYPE_PARAMETERS, TobConnection

LOGGER = logging.getLogger(__name__)


class AgentType(Enum):
    """
    Enumeration of supported agent types
    """
    issuer = "issuer"
    holder = "holder"
    verifier = "verifier"
    combined = "combined"


class AgentCfg:
    """
    Manage configuration settings for an Agent, including schemas bound for the ledger
    """
    def __init__(self, agent_type: str, wallet_id: str, **params):
        self.agent_id = params.get("id")
        try:
            self.agent_type = AgentType(agent_type)
        except KeyError:
            raise IndyConfigError("Unsupported agent type: {}".format(agent_type))
        self.cred_types = []
        self._instance = None
        self.opened = False
        self.registered = False
        self.synced = False
        self.wallet_id = wallet_id
        self.abbreviation = params.get("abbreviation")
        self.email = params.get("email")
        self.endpoint = params.get("endpoint")
        self.name = params.get("name")
        self.url = params.get("url")
        self.logo_b64 = params.get("logo_b64")
        self.logo_path = params.get("logo_path")
        self.link_secret_name = params.get("link_secret_name", "master-secret")

    @property
    def created(self) -> bool:
        """
        Accessor for the current created status of the agent instance
        """
        return self.instance is not None

    @property
    def did(self) -> str:
        """
        Accessor for DID of the agent's wallet
        """
        return self._instance and self._instance.did

    @property
    def extended_config(self) -> dict:
        """
        Accessor for the extended Agent configuration
        """
        ret = {}
        if self.endpoint:
            ret["endpoint"] = self.endpoint
        return ret

    @property
    def instance(self) -> _BaseAnchor:
        """
        Accessor for the current agent instance
        """
        return self._instance

    @property
    def role(self) -> str:
        """
        Accessor for the role of the agent to be registered on the ledger
        """
        return "TRUST_ANCHOR" if self.is_issuer else ""

    @property
    def status(self) -> dict:
        """
        Get the current status of the agent
        """
        return {
            "did": self.did,
            "created": self.created,
            "opened": self.opened,
            "registered": self.registered,
            "synced": self.synced,
        }

    @property
    def verkey(self) -> str:
        """
        Accessor for the verkey of the agent's wallet
        """
        return self._instance and self._instance.verkey

    async def create(self, wallet: 'WalletCfg', pool: NodePool) -> None:
        """
        Create the agent instance

        Args:
            wallet: the registered wallet configuration, previously created and opened
            pool: the initialized :class:`NodePool` instance for the wallet
        """
        if not self._instance:
            cls = None
            params = {"cfg": self.extended_config}
            if self.agent_type == AgentType.issuer:
                cls = BCRegistrarAnchor # combines Origin and Issuer
            elif self.agent_type == AgentType.holder:
                cls = HolderProver
            elif self.agent_type == AgentType.verifier:
                cls = Verifier
            elif self.agent_type == AgentType.combined:
                cls = OrgHubAnchor
            else:
                raise IndyConfigError("Unknown agent type")
            self._instance = cls(wallet.instance, pool, **params)
        await self.open()

    async def open(self) -> None:
        """
        Open the agent instance for storing or issuing credentials
        """
        if not self.opened:
            self.opened = await self._instance.open()
            if isinstance(self._instance, HolderProver):
                await self._instance.create_link_secret(self.link_secret_name)

    async def close(self) -> None:
        """
        Close the agent instance
        """
        if self.opened:
            await self._instance.close()
            self.opened = False

    async def get_endpoint(self, did: str) -> str:
        """
        Resolve a did to an endpoint
        """
        return await self._instance.get_endpoint(did)

    async def send_endpoint(self) -> None:
        """
        Write the agent's endpoint to the ledger.

        If end point is None, endpoint will be removed from ledger.
        """
        await self._instance.send_endpoint(self.endpoint)


    @property
    def is_holder(self):
        return self.agent_type == AgentType.holder or self.agent_type == AgentType.combined

    @property
    def is_issuer(self):
        return self.agent_type == AgentType.issuer or self.agent_type == AgentType.combined

    @property
    def is_verifier(self):
        return self.agent_type == AgentType.verifier or self.agent_type == AgentType.combined

    def add_credential_type(self, schema: 'SchemaCfg', **params) -> None:
        """
        Add a credential type to the Agent configuration

        Args:
            schema: the :class:`SchemaCfg` to be added
        """
        if not self.is_issuer:
            raise IndyConfigError("Only issuer agent may publish schemas")
        self.cred_types.append({
            "definition": schema,
            "ledger_schema": None,
            "cred_def": None,
            "params": params,
        })

    def find_credential_type(self, name: str, version: str, origin_did: str = None) -> dict:
        """
        Find the extended information for a specific schema, including the ledger schema
        definition and credential definition (if any)

        Args:
            name: the schema name to be located
            version: the schema version to be located
        """
        match = SchemaCfg(name, version, None, origin_did)
        for cred_type in self.cred_types:
            if cred_type["definition"].compare(match):
                return cred_type
        return None

    def get_connection_params(self, _connection: 'ConnectionCfg') -> dict:
        """
        Get parameters required for initializing the connection
        """
        if self.is_issuer:
            cred_specs = []
            for cred_type in self.cred_types:
                params = cred_type["params"]
                type_spec = {
                    "schema": cred_type["definition"],
                    "cred_def": cred_type["cred_def"],
                }
                for k in CRED_TYPE_PARAMETERS:
                    if k in params:
                        type_spec[k] = params[k]
                cred_specs.append(type_spec)
            return {
                "abbreviation": self.abbreviation,
                "credential_types": cred_specs,
                "did": self.did,
                "email": self.email,
                "name": self.name,
                "url": self.url,
                "logo_b64": self.logo_b64,
                "logo_path": self.logo_path,
            }
        return None


class ConnectionCfg:
    """
    Manage configuration settings for a connection between an issuer and a target
    """
    def __init__(self, connection_type: str, agent_id: str, agent_type: str, **params):
        self.connection_id = params.get("id")
        self.agent_id = agent_id
        self.agent_type = agent_type
        try:
            self.connection_type = ConnectionType(connection_type)
        except KeyError:
            raise IndyConfigError("Unsupported connection type: {}".format(connection_type))
        self._instance = None
        self.connection_params = params
        self.opened = False
        sign = params.get("sign_target", True)
        self.sign_target = sign and str(sign) != "0" and str(sign).lower() != "false"
        self.synced = False

        if self.connection_type != ConnectionType.TheOrgBook and \
                self.connection_type != ConnectionType.HTTP and \
                self.connection_type != ConnectionType.holder:
            raise IndyConfigError(
                "Only HTTP and internal Holder connections are currently supported")

    @property
    def created(self) -> bool:
        """
        Accessor for the current created status of the connection instance
        """
        return self._instance is not None

    @property
    def instance(self) -> ConnectionBase:
        """
        Accessor for the connection instance
        """
        return self._instance

    @property
    def status(self) -> dict:
        """
        Accessor for the status of the connection
        """
        return {
            "created": self.created,
            "opened": self.opened,
            "synced": self.synced,
        }

    async def create(self, agent_params: dict) -> None:
        """
        Create the connection instance

        Args:
            agent_params: extra parameters assembled by the agent service for this connection
        """
        if self.connection_type == ConnectionType.TheOrgBook:
            cls = TobConnection
        elif self.connection_type == ConnectionType.HTTP:
            cls = HttpConnection
        elif self.connection_type == ConnectionType.holder:
            cls = HolderConnection
        conn_params = self.connection_params.copy()
        conn_params["id"] = self.connection_id
        self._instance = cls(self.agent_id, self.agent_type, agent_params, conn_params)

    async def open(self, service: 'IndyService') -> None:
        """
        Open the connection

        Args:
            service: the Indy service handling this connection
        """
        if not self.opened:
            await self._instance.open(service)
            self.opened = True

    async def sync(self) -> None:
        """
        Perform synchronization of the connection instance
        """
        if not self.synced:
            await self._instance.sync()
            self.synced = True

    async def close(self) -> None:
        """
        Close the connection instance
        """
        if self.opened:
            await self._instance.close()
            self.opened = False


class ProofSpecCfg:
    """
    A proof request specification
    """
    def __init__(self, **params):
        self.spec_id = params.get("id")
        self.version = params.get("version")
        if not self.version:
            raise IndyConfigError("Missing version for proof spec: {}".format(self.spec_id))
        self.schemas = params.get("schemas")
        if not self.schemas:
            raise IndyConfigError("Missing schemas for proof spec: {}".format(self.spec_id))
        self.synced = not self.get_incomplete_schemas()

    @property
    def status(self) -> dict:
        """
        Accessor for the status of the proof specification
        """
        return {
            "synced": self.synced,
        }

    def get_incomplete_schemas(self) -> set:
        """
        Get a set of schemas which have yet to be populated with details from the ledger
        """
        missing = set()
        for schema in self.schemas:
            if not schema.get("definition"):
                s_key = schema["key"]
                missing.add((s_key["name"], s_key["version"], s_key.get("did")))
        return missing

    def populate_schema(self, found_schema: 'SchemaCfg') -> None:
        """
        Populate required schema details from the ledger
        """
        for schema in self.schemas:
            if not schema.get("definition"):
                s_key = schema["key"]
                cfg = SchemaCfg(s_key["name"], s_key["version"], None, s_key.get("did"))
                if cfg.compare(found_schema):
                    schema["definition"] = found_schema.copy()
                    if not schema.get("attributes"):
                        schema["attributes"] = found_schema.attr_names


class SchemaCfg:
    """
    A credential schema definition
    """
    def __init__(self, name: str, version: str = None, attributes=None, origin_did: str = None):
        self.name = name
        self.version = version
        self._attributes = []
        if attributes:
            self.attributes = attributes
        self.origin_did = origin_did

    @property
    def schema_id(self) -> str:
        """
        Accessor for the schema_id of this schema
        """
        return schema_id(self.origin_did, self.name, self.version)

    @property
    def attributes(self) -> list:
        """
        Accessor for the extended schema attributes list

        Returns:
            a copy of the schema attributes
        """
        return self._attributes.copy()

    @attributes.setter
    def attributes(self, value) -> None:
        """
        Setter for the schema attributes list
        """
        self._attributes = []
        if isinstance(value, Mapping):
            for name, attr in value.items():
                self.add_attribute(attr, name)
        elif isinstance(value, Sequence):
            for attr in value:
                self.add_attribute(attr)
        else:
            raise IndyConfigError('Unsupported type for attributes: {}'.format(value))

    @property
    def attr_names(self) -> list:
        """
        Accessor for the schema attribute names

        Returns:
            the attribute names only
        """
        return tuple(attr['name'] for attr in self._attributes)

    def add_attribute(self, attr, name=None) -> None:
        """
        Add an attribute to the schema including optional type information

        Args:
            attr: a dict or str representing the attribute
            name: the name of the attribute
        """
        if isinstance(attr, Mapping):
            if name is not None:
                attr['name'] = name
            self._attributes.append(attr)
        elif isinstance(attr, str):
            attr = {'name': attr}
            self._attributes.append(attr)
        elif attr is None and name:
            self._attributes.append({'name': name})
        else:
            raise IndyConfigError('Unsupported type for attribute: {}'.format(attr))

    def copy(self) -> 'SchemaCfg':
        """
        Create a copy of this :class:`SchemaCfg` instance
        """
        return SchemaCfg(self.name, self.version, self._attributes, self.origin_did)

    def validate(self, value) -> None:
        """
        Perform validation of a set of attribute values against the schema
        """
        pass

    def compare(self, schema: 'SchemaCfg') -> bool:
        """
        Check whether this schema instance and another are compatible.
        Note: schemas with an empty issuer DID will match schemas with a blank issuer DID,
        or the same DID
        """
        if self.name != schema.name:
            return False
        if self.version and schema.version and self.version != schema.version:
            return False
        if self.origin_did and schema.origin_did and self.origin_did != schema.origin_did:
            return False
        if self.attributes and schema.attributes and self.attributes != schema.attributes:
            return False
        return True

    def __repr__(self) -> str:
        return 'SchemaCfg(name={}, version={}, origin_did={})'.format(
            self.name, self.version, self.origin_did)


class SchemaManager:
    """
    A manager class for handling a set of loaded credential schema definitions
    """

    def __init__(self):
        self._schemas = []

    @property
    def schemas(self) -> list:
        """
        An accessor for the list of all loaded schemas
        """
        return self._schemas.copy()

    def add_schema(self, schema, override=False) -> None:
        """
        Add a schema to the manager

        Args:
            schema: a :class:`SchemaCfg` or dict instance
            override: replace an existing schema if any
        """
        if not isinstance(schema, SchemaCfg):
            if not isinstance(schema, Mapping):
                raise IndyConfigError('Unsupported type for schema: {}'.format(schema))
            name = schema.get('name')
            if not name:
                raise IndyConfigError('Missing schema name')
            schema = SchemaCfg(name, schema.get('version'), schema.get('attributes'))
        found = self.find(schema.name, schema.version)
        if found:
            if override:
                self.remove_schema(found)
            else:
                raise IndyConfigError('Duplicate schema definition: {}'.format(schema))
        self._schemas.append(schema)

    def remove_schema(self, schema, version=None) -> None:
        """
        Remove an existing schema from the manager

        Args:
            schema: the schema name
            version: the schema version
        """
        if isinstance(schema, str):
            schema = self.find(schema, version)
        self._schemas.remove(schema)

    def load(self, values: Sequence, override=False) -> None:
        """
        Load a list of schemas and add each to the manager

        Args:
            values: the list of schema definitions
            override: replace existing defined schemas of the same name and version
        """
        for spec in values:
            self.add_schema(spec, override)

    def find(self, name: str, version: str = None) -> SchemaCfg:
        """
        Locate a defined schema

        Args:
            name: the schema name
            version: the schema version

        Returns:
            the located :class:`SchemaCfg` instance, if any
        """
        found = None
        for schema in self._schemas:
            if schema.name == name:
                if version is not None:
                    if schema.version == version:
                        found = schema
                        break
                else:
                    if found is None or LooseVersion(found.version) < LooseVersion(schema.version):
                        found = schema
        return found


class WalletCfg:
    """
    Manage configuration settings for an Indy wallet
    """
    def __init__(self, **params):
        self.wallet_id = params.get("id")
        self.name = params.get("name")
        if not self.name:
            raise IndyConfigError("Missing wallet name")
        self.seed = params.get("seed")
        if not self.seed:
            raise IndyConfigError("Missing seed for wallet '{}'".format(self.name))
        seed_valid = False
        if len(self.seed) == 32:
            seed_valid = True
        elif self.seed[-1:] == "=":
            try:
                decoded = base64.b64decode(bytes(self.seed, 'ascii'))
                if len(decoded) == 32:
                    seed_valid = True
            except binascii.Error:
                pass
        if not seed_valid:
            raise IndyConfigError(
                "Wallet seed length is not 32 characters and/or not valid base64: {}".format(
                self.seed)
            )
        self.type = params.get("type")
        self.params = params.get("params") or {}
        if "freshness_time" not in self.params:
            self.params["freshness_time"] = 0
        self.access_creds = params.get("access_creds")
        self._instance = None

    @property
    def created(self) -> bool:
        """
        Accessor for the current created status of the wallet instance
        """
        return self._instance and self._instance.created

    @property
    def instance(self) -> Wallet:
        """
        Accessor for the wallet instance
        """
        return self._instance

    @property
    def opened(self) -> bool:
        """
        Accessor for the current opened status of the wallet instance
        """
        return self._instance and self._instance.handle is not None

    @property
    def status(self) -> dict:
        """
        Accessor for the current status of the wallet instance
        """
        return {
            "created": self.created,
            "opened": self.opened,
        }

    async def load_storage_library(self, storage_type):
        # load storage library for postgres
        if storage_type == "postgres":
            try:
                await register_wallet_storage_library(storage_type,"libindystrgpostgres.so","postgreswallet_fn_")
            except IndyError as x_indy:
                if x_indy.error_code == ErrorCode.WalletTypeAlreadyRegisteredError:
                    LOGGER.info('Wallet already exists: %s', self.name)
                else:
                    # ignore errors, the error will occur on creating or opening the wallet
                    LOGGER.error('Wallet.register <!< indy error on load of wallet storage')
                    raise

    async def create(self) -> None:
        """
        Create the wallet instance
        """
        await self.load_storage_library(self.type)
        self._instance = Wallet(
            self.seed,
            self.name,
            self.type,
            self.params,
            self.access_creds)
        await self.instance.create()

    async def open(self):
        """
        Open the wallet instance
        """
        await self._instance.open()

    async def close(self) -> None:
        """
        Close the wallet instance
        """
        if self.opened:
            await self._instance.close()



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Handle rendering of issue-credential forms
"""


import logging

from aiohttp import web
import aiohttp_jinja2

from ..common.util import log_json, normalize_credential_ids
from ..indy.errors import IndyClientError

LOGGER = logging.getLogger(__name__)


async def render_form(form: dict, request: web.Request) -> web.Response:
    """
    Render a form definition by:

        - performing a proof request if needed,
        - collecting any values needed from the proof request,
        - adding any user input,
        - populating template variables,
        - and rendering the template defined by the form definition

    Args:
        form: The form definition
        request: The request received by aiohttp
    """
    #pylint: disable=broad-except
    proof_meta = form.get("proof_request")
    proof_response = None
    service_mgr = request.app["manager"]

    if proof_meta:
        try:
            client = service_mgr.get_client()
            proof_req = await client.generate_proof_request(proof_meta["id"])

            params = {}
            if "params" in proof_meta:
                for attr_name, param in proof_meta["params"].items():
                    if isinstance(param, str):
                        param_from = param
                    elif isinstance(param, dict):
                        param_from = param.get("from")
                    if param_from:
                        val = request.query.get(param_from)
                        if val is not None and val != '':
                            params[attr_name] = val

            cred_ids = request.query.get("credential_ids", request.query.get("credential_id"))
            cred_ids = normalize_credential_ids(cred_ids)

            verified = await client.request_proof(
                proof_meta["connection_id"], proof_req, cred_ids, params)
            proof_response = {
                "success": True,
                "verified": verified.verified == "true",
                "parsed_proof": verified.parsed_proof,
                "proof": verified.proof.proof,
            }
        except IndyClientError as e:
            proof_response = {"success": False, "result": str(e)}
        log_json("Proof response:", proof_response, LOGGER, logging.INFO)

    tpl_name = form.get("template", "index.html")
    tpl_vars = {
        "inputs": {},
        "request": {},
        "proof_response": proof_response,
        "THE_ORG_BOOK_APP_URL": service_mgr.env.get("TOB_APP_URL")
    }
    tpl_vars["inputs"].update(request.query)
    tpl_vars["request"].update(request.query)
    if proof_response and proof_response["success"]:
        # currently flattening attributes from different schemas
        proof_attrs = {}
        for attrs in proof_response["parsed_proof"].values():
            proof_attrs.update(attrs)

        if "inputs" in proof_req:
            for input_name, claim_name in proof_req["inputs"].items():
                tpl_vars["inputs"][input_name] = proof_attrs.get(claim_name, "")
        else:
            tpl_vars["inputs"].update(proof_attrs)
    tpl_vars.update(form)
    if "hidden" not in tpl_vars:
        tpl_vars["hidden"] = []
    if "connection_id" not in tpl_vars["hidden"]:
        tpl_vars["hidden"].append("connection_id")
    tpl_vars["inputs"]["connection_id"] = form.get("connection_id", "")
    tpl_vars["path"] = request.rel_url

    return aiohttp_jinja2.render_template(tpl_name, request, tpl_vars)



#!/usr/bin/env python3
#
# Copyright 2017-2018 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import aiohttp
import asyncio
import async_timeout

URL='http://localhost:5000/health'
BATCH=100
RUNS=1

async def fetch(session, url):
    async with async_timeout.timeout(30):
        async with session.get(url) as response:
            return response.status

async def main():
    async with aiohttp.ClientSession() as session:
        tasks = [
            asyncio.ensure_future(fetch(session, URL))
            for i in range(BATCH)
        ]
        responses = await asyncio.gather(*tasks)
        print(responses)

loop = asyncio.get_event_loop()
loop.run_until_complete(main())



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Connection handling specific to using TheOrgBook as a holder/prover
"""

import base64
import logging
import pathlib

from .connection import HttpConnection, HttpSession
from .errors import IndyConfigError, IndyConnectionError
from ..common.util import log_json

LOGGER = logging.getLogger(__name__)

CRED_TYPE_PARAMETERS = (
    "cardinality_fields",
    "credential",
    "description",
    "issuer_url",
    "mapping",
    "topic",
    "logo_b64",
    "logo_path",
    "visible_fields",
)


def encode_logo_image(config: dict, path_root: str) -> str:
    """
    Encode logo image as base64 for transmission
    """
    if config.get("logo_b64"):
        return config["logo_b64"]
    elif config.get("logo_path"):
        path = pathlib.Path(path_root, config["logo_path"])
        if path.is_file():
            content = path.read_bytes()
            if content:
                return base64.b64encode(content).decode("ascii")
        else:
            LOGGER.warning("No file found at logo path: %s", path)
    return None


def assemble_issuer_spec(config: dict) -> dict:
    """
    Create the issuer JSON definition which will be submitted to TheOrgBook
    """
    issuer_spec = {}
    issuer_email = config.get("email")
    if not issuer_email:
        raise IndyConfigError("Missing issuer email address")
    issuer_did = config.get("did")
    if not issuer_did:
        raise IndyConfigError("Missing issuer DID")

    config_root = config.get("config_root", ".")
    issuer_spec["issuer"] = {
        "did": issuer_did,
        "name": config.get("name") or "",
        "abbreviation": config.get("abbreviation") or "",
        "email": issuer_email,
        "url": config.get("url") or "",
        "logo_b64": encode_logo_image(config, config_root),
    }

    if not issuer_spec["issuer"]["name"]:
        raise IndyConfigError("Missing issuer name")

    cred_type_specs = config.get("credential_types")
    if not cred_type_specs:
        raise IndyConfigError("Missing credential_types")
    ctypes = []
    for type_spec in cred_type_specs:
        schema = type_spec["schema"]
        if not type_spec.get("topic"):
            raise IndyConfigError("Missing 'topic' for credential type")
        ctype = {
            "name": type_spec.get("description") or schema.name,
            "endpoint": type_spec.get("issuer_url") or issuer_spec["issuer"]["url"],
            "schema": schema.name,
            "version": schema.version,
            "topic": type_spec["topic"],
            "credential_def_id": type_spec["cred_def"]["id"],
        }
        for k in CRED_TYPE_PARAMETERS:
            if k in type_spec and k not in ctype:
                ctype[k] = type_spec[k]
        ctype["logo_b64"] = encode_logo_image(type_spec, config_root)
        if "logo_path" in ctype:
            del ctype["logo_path"]
        ctypes.append(ctype)
    issuer_spec["credential_types"] = ctypes
    return issuer_spec


class TobConnection(HttpConnection):
    """
    A class for managing communication with TheOrgBook API and performing the initial
    synchronization as an issuer
    """

    async def sync(self) -> None:
        """
        Submit the issuer JSON definition to TheOrgBook to register our service
        """
        if self.agent_type == "issuer":
            spec = assemble_issuer_spec(self.agent_params)
            log_json("Issuer spec:", spec, LOGGER)
            response = await self.post_json(
                "indy/register-issuer", spec
            )
            result = response.get("result")
            if not response.get("success"):
                raise IndyConnectionError(
                    "Issuer service was not registered: {}".format(result),
                    400,
                    response,
                )

    @property
    def path_prefix(self):
        return "indy/"

    async def fetch_list(self, path: str) -> dict:
        """
        A standard request to a `list`-style API method

        Args:
            path: The relative path to the API method
        """
        url = self.get_api_url(path)
        LOGGER.debug("fetch_list: %s", url)
        async with HttpSession("fetch_list", self._http_client) as handler:
            response = await handler.client.get(url)
            await handler.check_status(response)
            return await response.json()
